# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xtrack as xt
import xpart as xp
from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_aperture_turn_ele_and_monitor(test_context):
    x_aper_min = -0.1
    x_aper_max = 0.2
    y_aper_min = 0.2
    y_aper_max = 0.3

    part_gen_range = 0.35
    n_part=100

    particles = xp.Particles(_context=test_context,
                             p0c=6500e9,
                             x=np.zeros(n_part),
                             px=np.linspace(-1, 1, n_part),
                             y=np.zeros(n_part),
                             py=np.linspace(-2, 2, n_part),
                            )

    # Build a small test line
    tot_length = 2.
    n_slices = 10000
    line = xt.Line(elements=n_slices*[xt.Drift(
                                            length=tot_length/n_slices)],
                    element_names=['drift{ii}' for ii in range(n_slices)])

    line.build_tracker(_context=test_context)
    n_turns = 3
    line.track(particles, num_turns=n_turns, turn_by_turn_monitor=True)

    part_id = test_context.nparray_from_context_array(particles.particle_id)
    part_state = test_context.nparray_from_context_array(particles.state)
    part_x = test_context.nparray_from_context_array(particles.x)
    part_y = test_context.nparray_from_context_array(particles.y)
    part_px = test_context.nparray_from_context_array(particles.px)
    part_py = test_context.nparray_from_context_array(particles.py)
    part_s = test_context.nparray_from_context_array(particles.s)
    part_at_turn = test_context.nparray_from_context_array(particles.at_turn)
    part_at_element = test_context.nparray_from_context_array(particles.at_element)

    id_alive = part_id[part_state>0]

    #x = px*s
    s_expected = []
    s_tot = tot_length*n_turns
    global_aper_limit = line.config.XTRACK_GLOBAL_XY_LIMIT
    for ii in range(n_part):
        if np.abs(part_px[ii]) * s_tot > global_aper_limit:
            s_expected_x = np.abs(global_aper_limit / part_px[ii])
        else:
            s_expected_x = s_tot

        if np.abs(part_py[ii] * s_tot) > global_aper_limit:
            s_expected_y = np.abs(global_aper_limit / part_py[ii])
        else:
            s_expected_y = s_tot

        if s_expected_x<s_expected_y:
            s_expected.append(s_expected_x)
        else:
            s_expected.append(s_expected_y)

    s_expected = np.array(s_expected)
    at_turn_expected = np.int_(np.clip(np.floor(s_expected/tot_length),
                                       0, n_turns))
    at_element_expected = np.floor((s_expected-tot_length*at_turn_expected)
                                         /(tot_length/n_slices)) + 1
    at_element_expected = np.int_(np.clip(at_element_expected, 0,
                                          n_slices-1))

    xo.assert_allclose(part_s + at_turn_expected*line.get_length(), s_expected,
                       atol=1e-3)
    xo.assert_allclose(at_turn_expected, part_at_turn)

    # I need to add a tolerance of one element as a mismatch is visible
    # on a few slices due to rounding
    xo.assert_allclose(at_element_expected, part_at_element, atol=1.1)

    # Test monitor
    mon = line.record_last_track
    for ii in range(n_part):
        iidd = part_id[ii]
        this_at_turn = part_at_turn[ii]
        this_px = part_px[ii]
        this_py = part_py[ii]

        for tt in range(n_turns):
            if tt<=this_at_turn:
                assert(mon.at_turn[iidd, tt] == tt)
                assert(np.isclose(mon.s[iidd, tt], 0., atol=1e-14))
                assert(np.isclose(mon.x[iidd, tt], tt*tot_length*this_px,
                                  atol=1e-14))
                assert(np.isclose(mon.y[iidd, tt], tt*tot_length*this_py,
                                  atol=1e-14))
                assert(np.isclose(mon.px[iidd, tt], this_px, atol=1e-14))
                assert(np.isclose(mon.py[iidd, tt], this_py, atol=1e-14))
            else:
                assert(mon.at_turn[iidd, tt] == 0)
                assert(mon.s[iidd, tt] == 0)
                assert(mon.x[iidd, tt] == 0)
                assert(mon.y[iidd, tt] == 0)
                assert(mon.px[iidd, tt] == 0)
                assert(mon.py[iidd, tt] == 0)


@for_all_test_contexts
def test_custom_monitor(test_context):
    x_aper_min = -0.1
    x_aper_max = 0.2
    y_aper_min = 0.2
    y_aper_max = 0.3

    part_gen_range = 0.35
    n_part=100

    particles = xp.Particles(_context=test_context,
                             p0c=6500e9,
                             x=np.zeros(n_part),
                             px=np.linspace(-1, 1, n_part),
                             y=np.zeros(n_part),
                             py=np.linspace(-2, 2, n_part),
                            )

    # Build a small test line
    tot_length = 2.
    n_slices = 10000
    line = xt.Line(elements=n_slices*[xt.Drift(
                                            length=tot_length/n_slices)],
                    element_names=['drift{ii}' for ii in range(n_slices)])

    line.build_tracker(_context=test_context)
    n_turns = 6
    start_monitor_at_turn = 2
    part_id_monitor_start = 5
    part_id_monitor_end = n_part - 7

    monitor = xt.ParticlesMonitor(_context=test_context,
            start_at_turn=start_monitor_at_turn,
            stop_at_turn=n_turns,
            particle_id_range=(part_id_monitor_start, part_id_monitor_end))
    for _ in range(n_turns):
        line.track(particles, turn_by_turn_monitor=monitor)

    part_id = test_context.nparray_from_context_array(particles.particle_id)
    part_state = test_context.nparray_from_context_array(particles.state)
    part_x = test_context.nparray_from_context_array(particles.x)
    part_y = test_context.nparray_from_context_array(particles.y)
    part_px = test_context.nparray_from_context_array(particles.px)
    part_py = test_context.nparray_from_context_array(particles.py)
    part_s = test_context.nparray_from_context_array(particles.s)
    part_at_turn = test_context.nparray_from_context_array(particles.at_turn)
    part_at_element = test_context.nparray_from_context_array(particles.at_element)

    id_alive = part_id[part_state>0]

    #x = px*s
    s_expected = []
    s_tot = tot_length*n_turns
    global_aper_limit = line.config.XTRACK_GLOBAL_XY_LIMIT
    for ii in range(n_part):
        if np.abs(part_px[ii]) * s_tot > global_aper_limit:
            s_expected_x = np.abs(global_aper_limit / part_px[ii])
        else:
            s_expected_x = s_tot

        if np.abs(part_py[ii] * s_tot) > global_aper_limit:
            s_expected_y = np.abs(global_aper_limit / part_py[ii])
        else:
            s_expected_y = s_tot

        if s_expected_x<s_expected_y:
            s_expected.append(s_expected_x)
        else:
            s_expected.append(s_expected_y)

    s_expected = np.array(s_expected)
    at_turn_expected = np.int_(np.clip(np.floor(s_expected/tot_length),
                                       0, n_turns))
    at_element_expected = np.floor((s_expected-tot_length*at_turn_expected)
                                         /(tot_length/n_slices)) + 1
    at_element_expected = np.int_(np.clip(at_element_expected, 0,
                                          n_slices-1))

    xo.assert_allclose(part_s + at_turn_expected * line.get_length(),
                       s_expected, atol=1e-3)
    xo.assert_allclose(at_turn_expected, part_at_turn)

    # I need to add a tolerance of one element as a mismatch is visible
    # on a few slices due to rounding
    xo.assert_allclose(at_element_expected, part_at_element, atol=1.1)

    # Test monitor
    mon = monitor
    for ii in range(n_part):
        iidd = part_id[ii]
        this_at_turn = part_at_turn[ii]
        this_px = part_px[ii]
        this_py = part_py[ii]

        if iidd < part_id_monitor_start or iidd >= part_id_monitor_end:
            continue

        for tt in range(n_turns):
            if tt < start_monitor_at_turn:
                continue
            if tt<=this_at_turn:
                assert(mon.at_turn[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn] == tt)
                assert(mon.particle_id[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn] == iidd)
                assert(np.isclose(mon.s[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn],
                    0., atol=1e-14))
                assert(np.isclose(mon.x[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn],
                    tt*tot_length*this_px,
                    atol=1e-14))
                assert(np.isclose(mon.y[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn],
                    tt*tot_length*this_py,
                    atol=1e-14))
                assert(np.isclose(mon.px[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn],
                    this_px, atol=1e-14))
                assert(np.isclose(mon.py[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn],
                    this_py, atol=1e-14))
            else:
                assert(mon.at_turn[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn] == 0)
                assert(mon.particle_id[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn] == 0)
                assert(mon.s[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn] == 0)
                assert(mon.x[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn] == 0)
                assert(mon.y[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn] == 0)
                assert(mon.px[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn] == 0)
                assert(mon.py[iidd - part_id_monitor_start,
                    tt - start_monitor_at_turn] == 0)



import xtrack as xt
import numpy as np

to_test = 'line'
to_test = 'env'

env = xt.Environment()

env.vars({
    'k.1': 1.,
    'a': 2.,
    'b': '2 * a + k.1',
})

line = env.new_line([])

ee = {'env': env, 'line': line}[to_test]

assert ee.vv['b'] == 2 * 2 + 1

ee.vars['a'] = ee.vars['k.1']
assert ee.vv['b'] == 2 * 1 + 1

ee.vars(a=3.)
ee.vars({'k.1': 'a'})
assert ee.vv['k.1'] == 3.
assert ee.vv['b'] == 2 * 3 + 3.

ee.vars['k.1'] = 2 * ee.vars['a'] + 5
assert ee.vv['k.1'] == 2 * 3 + 5
assert ee.vv['b'] == 2 * 3 + 2 * 3 + 5

ee.vars.set('a', 4.)
assert ee.vv['k.1'] == 2 * 4 + 5
assert ee.vv['b'] == 2 * 4 + 2 * 4 + 5

ee.vars.set('k.1', '2*a + 5')
assert ee.vv['k.1'] == 2 * 4 + 5
assert ee.vv['b'] == 2 * 4 + 2 * 4 + 5

ee.vars.set('k.1', 3 * ee.vars['a'] + 6)
assert ee.vv['k.1'] == 3 * 4 + 6
assert ee.vv['b'] == 2 * 4 + 3 * 4 + 6

ee.set('a', 0.)
assert ee.vv['k.1'] == 3 * 0 + 6
assert ee.vv['b'] == 2 * 0 + 3 * 0 + 6

ee.set('a', 2.)
ee.set('k.1', '2 * a + 5')
assert ee.vv['k.1'] == 2 * 2 + 5
assert ee.vv['b'] == 2 * 2 + 2 * 2 + 5

ee.set('k.1', 3 * ee.vars['a'] + 6)
assert ee.vv['k.1'] == 3 * 2 + 6
assert ee.vv['b'] == 2 * 2 + 3 * 2 + 6

assert hasattr(ee.ref['k.1'], '_value') # is a Ref

ee.ref['a'] = 0
assert ee.vv['k.1'] == 3 * 0 + 6
assert ee.vv['b'] == 2 * 0 + 3 * 0 + 6

ee.ref['a'] = 2
ee.ref['k.1'] = 2 * ee.ref['a'] + 5
assert ee.vv['k.1'] == 2 * 2 + 5
assert ee.vv['b'] == 2 * 2 + 2 * 2 + 5

#--------------------------------------------------

ee.vars({
    'a': 4.,
    'b': '2 * a + 5',
    'k.1': '2 * a + 5',
})

env.new('bb', xt.Bend, k0='2 * b', length=3+env.vars['a'] + env.vars['b'],
        h=5.)
assert env['bb'].k0 == 2 * (2 * 4 + 5)
assert env['bb'].length == 3 + 4 + 2 * 4 + 5
assert env['bb'].h == 5.

env.vars['a'] = 2.
assert env['bb'].k0 == 2 * (2 * 2 + 5)
assert env['bb'].length == 3 + 2 + 2 * 2 + 5
assert env['bb'].h == 5.

line = env.new_line([
    env.new('bb1', 'bb', length=3*env.vars['a'], at='2*a'),
    env.place('bb', at=10 * env.vars['a'], from_='bb1'),
])

assert hasattr(env.ref['bb1'].length, '_value') # is a Ref
assert not hasattr(env['bb1'].length, '_value') # a number
assert env.ref['bb1'].length._value == 3 * 2
assert env['bb1'].length == 3 * 2

assert hasattr(env.ref['bb1'].length, '_value') # is a Ref
assert not hasattr(env['bb1'].length, '_value') # a number
assert env.ref['bb1'].length._value == 3 * 2
assert env['bb1'].length == 3 * 2

assert line['bb1'] is not env['bb']
assert line['bb'] is env['bb']

a = env.vv['a']
assert line['bb1'].length == 3 * a
assert line['bb1'].k0 == 2 * (2 * a + 5)
assert line['bb1'].h == 5.

assert line['bb'].k0 == 2 * (2 * a + 5)
assert line['bb'].length == 3 + a + 2 * a + 5
assert line['bb'].h == 5.

tt = line.get_table(attr=True)
tt['s_center'] = tt['s'] + tt['length']/2

assert np.all(tt.name ==  np.array(['drift_1', 'bb1', 'drift_2', 'bb', '_end_point']))

assert tt['s_center', 'bb1'] == 2*a
assert tt['s_center', 'bb'] - tt['s_center', 'bb1'] == 10*a

old_a = a
line.vars['a'] = 3.
a = line.vv['a']
assert line['bb1'].length == 3 * a
assert line['bb1'].k0 == 2 * (2 * a + 5)
assert line['bb1'].h == 5.

assert line['bb'].k0 == 2 * (2 * a + 5)
assert line['bb'].length == 3 + a + 2 * a + 5
assert line['bb'].h == 5.

tt_new = line.get_table(attr=True)

# Drifts are not changed:
tt_new['length', 'drift_1'] == tt['length', 'drift_1']
tt_new['length', 'drift_2'] == tt['length', 'drift_2']




import xtrack as xt
import xobjects as xo
import numpy as np

assert_allclose = xo.assert_allclose

elements = {
    'e0': xt.Bend(k0=0.3, h=0.31, length=1),
    'e1': xt.Replica(parent_name='e0'),
    'e2': xt.Bend(k0=-0.4, h=-0.41, length=1),
    'e3': xt.Replica(parent_name='e2'),
    'e4': xt.Replica(parent_name='e3'), # Replica of replica
}
line = xt.Line(elements=elements,
               element_names=list(elements.keys()))
line.build_tracker()

element_no_repl={
    'e0': xt.Bend(k0=0.3, h=0.31, length=1),
    'e1': xt.Bend(k0=0.3, h=0.31, length=1),
    'e2': xt.Bend(k0=-0.4, h=-0.41, length=1),
    'e3': xt.Bend(k0=-0.4, h=-0.41, length=1),
    'e4': xt.Bend(k0=-0.4, h=-0.41, length=1),
}

line_no_repl = xt.Line(elements=element_no_repl,
                       element_names=list(element_no_repl.keys()))
line_no_repl.build_tracker()

tt = line.get_table()
tt_no_repl = line_no_repl.get_table()

assert np.all(tt.name == ['e0', 'e1', 'e2', 'e3', 'e4', '_end_point'])
assert np.all(tt_no_repl.name == ['e0', 'e1', 'e2', 'e3', 'e4', '_end_point'])

assert np.all(tt.parent_name == [None, 'e0', None, 'e2', 'e3', None])
assert np.all(tt_no_repl.parent_name == [None, None, None, None, None, None])

assert_allclose(tt.s, [0., 1., 2., 3., 4., 5.], rtol=0, atol=1e-14)
assert_allclose(tt_no_repl.s, [0., 1., 2., 3., 4., 5.], rtol=0, atol=1e-14)

p0 = xt.Particles(p0c=1e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03)
p1 = p0.copy()
p2 = p0.copy()

line.track(p1)
line_no_repl.track(p2)

assert_allclose(p2.x, p1.x, rtol=0, atol=1e-14)
assert_allclose(p2.px, p1.px, rtol=0, atol=1e-14)
assert_allclose(p2.y, p1.y, rtol=0, atol=1e-14)
assert_allclose(p2.py, p1.py, rtol=0, atol=1e-14)
assert_allclose(p2.zeta, p1.zeta, rtol=0, atol=1e-14)
assert_allclose(p2.delta, p1.delta, rtol=0, atol=1e-14)

line.slice_thick_elements(
    slicing_strategies=[
        xt.Strategy(None),
        xt.Strategy(xt.Teapot(3, mode='thick'), name='e2|e3|e4')])
line.build_tracker()

line_no_repl.slice_thick_elements(
    slicing_strategies=[
        xt.Strategy(None),
        xt.Strategy(xt.Teapot(3, mode='thick'), name='e2|e3|e4')])
line_no_repl.build_tracker()

tt = line.get_table()
tt_no_repl = line_no_repl.get_table()

assert np.all(tt.name == np.array([
       'e0', 'e1', 'e2_entry', 'e2..entry_map', 'e2..0', 'e2..1', 'e2..2',
       'e2..exit_map', 'e2_exit', 'e3_entry', 'e3..entry_map', 'e3..0',
       'e3..1', 'e3..2', 'e3..exit_map', 'e3_exit', 'e4_entry',
       'e4..entry_map', 'e4..0', 'e4..1', 'e4..2', 'e4..exit_map',
       'e4_exit', '_end_point']))

assert np.all(tt_no_repl.name == np.array([
       'e0', 'e1', 'e2_entry', 'e2..entry_map', 'e2..0', 'e2..1', 'e2..2',
       'e2..exit_map', 'e2_exit', 'e3_entry', 'e3..entry_map', 'e3..0',
       'e3..1', 'e3..2', 'e3..exit_map', 'e3_exit', 'e4_entry',
       'e4..entry_map', 'e4..0', 'e4..1', 'e4..2', 'e4..exit_map',
       'e4_exit', '_end_point']))

assert np.all(tt.parent_name == np.array([
       None, 'e0', None, 'e2', 'e2', 'e2', 'e2', 'e2', None, None, 'e2',
       'e2', 'e2', 'e2', 'e2', None, None, 'e2', 'e2', 'e2', 'e2', 'e2',
       None, None]))
assert np.all(tt_no_repl.parent_name == np.array([
       None, None, None, 'e2', 'e2', 'e2', 'e2', 'e2', None, None, 'e3',
       'e3', 'e3', 'e3', 'e3', None, None, 'e4', 'e4', 'e4', 'e4', 'e4',
       None, None]))

assert_allclose(tt.s, np.array([
       0.        , 1.        , 2.        , 2.        , 2.        ,
       2.16666667, 2.83333333, 3.        , 3.        , 3.        ,
       3.        , 3.        , 3.16666667, 3.83333333, 4.        ,
       4.        , 4.        , 4.        , 4.        , 4.16666667,
       4.83333333, 5.        , 5.        , 5.        ]))

assert_allclose(tt_no_repl.s, np.array([
       0.        , 1.        , 2.        , 2.        , 2.        ,
       2.16666667, 2.83333333, 3.        , 3.        , 3.        ,
       3.        , 3.        , 3.16666667, 3.83333333, 4.        ,
       4.        , 4.        , 4.        , 4.        , 4.16666667,
       4.83333333, 5.        , 5.        , 5.        ]))


p1 = p0.copy()
p2 = p0.copy()

line.track(p1)
line_no_repl.track(p2)

assert_allclose(p2.x, p1.x, rtol=0, atol=1e-14)
assert_allclose(p2.px, p1.px, rtol=0, atol=1e-14)
assert_allclose(p2.y, p1.y, rtol=0, atol=1e-14)
assert_allclose(p2.py, p1.py, rtol=0, atol=1e-14)
assert_allclose(p2.zeta, p1.zeta, rtol=0, atol=1e-14)
assert_allclose(p2.delta, p1.delta, rtol=0, atol=1e-14)

line.discard_tracker()
line.insert_element(name='mkins1', element=xt.Marker(), at_s=0.5)
line.insert_element(name='mkins2', element=xt.Marker(), at_s=1.5)
line.insert_element(name='mkins3', element=xt.Marker(), at_s=2.5)
line.insert_element(name='mkins4', element=xt.Marker(), at_s=3.5)
line.insert_element(name='mkins5', element=xt.Marker(), at_s=4.5)
line.build_tracker()

line_no_repl.discard_tracker()
line_no_repl.insert_element(name='mkins1', element=xt.Marker(), at_s=0.5)
line_no_repl.insert_element(name='mkins2', element=xt.Marker(), at_s=1.5)
line_no_repl.insert_element(name='mkins3', element=xt.Marker(), at_s=2.5)
line_no_repl.insert_element(name='mkins4', element=xt.Marker(), at_s=3.5)
line_no_repl.insert_element(name='mkins5', element=xt.Marker(), at_s=4.5)
line_no_repl.build_tracker()

tt = line.get_table()
tt_no_repl = line_no_repl.get_table()


assert np.all(tt.name == np.array([
       'e0_entry', 'e0..entry_map', 'e0..0', 'mkins1', 'e0..1',
       'e0..exit_map', 'e0_exit', 'e1_entry', 'e1..entry_map', 'e1..0',
       'mkins2', 'e1..1', 'e1..exit_map', 'e1_exit', 'e2_entry',
       'e2..entry_map', 'e2..0', 'e2..1_entry', 'e2..1..0', 'mkins3',
       'e2..1..1', 'e2..1_exit', 'e2..2', 'e2..exit_map', 'e2_exit',
       'e3_entry', 'e3..entry_map', 'e3..0', 'e3..1_entry', 'e3..1..0',
       'mkins4', 'e3..1..1', 'e3..1_exit', 'e3..2', 'e3..exit_map',
       'e3_exit', 'e4_entry', 'e4..entry_map', 'e4..0', 'e4..1_entry',
       'e4..1..0', 'mkins5', 'e4..1..1', 'e4..1_exit', 'e4..2',
       'e4..exit_map', 'e4_exit', '_end_point']))

assert np.all(tt_no_repl.name == np.array([
       'e0_entry', 'e0..entry_map', 'e0..0', 'mkins1', 'e0..1',
       'e0..exit_map', 'e0_exit', 'e1_entry', 'e1..entry_map', 'e1..0',
       'mkins2', 'e1..1', 'e1..exit_map', 'e1_exit', 'e2_entry',
       'e2..entry_map', 'e2..0', 'e2..1_entry', 'e2..1..0', 'mkins3',
       'e2..1..1', 'e2..1_exit', 'e2..2', 'e2..exit_map', 'e2_exit',
       'e3_entry', 'e3..entry_map', 'e3..0', 'e3..1_entry', 'e3..1..0',
       'mkins4', 'e3..1..1', 'e3..1_exit', 'e3..2', 'e3..exit_map',
       'e3_exit', 'e4_entry', 'e4..entry_map', 'e4..0', 'e4..1_entry',
       'e4..1..0', 'mkins5', 'e4..1..1', 'e4..1_exit', 'e4..2',
       'e4..exit_map', 'e4_exit', '_end_point']))

assert np.all(tt.parent_name == np.array([
       None, 'e0', 'e0', None, 'e0', 'e0', None, None, 'e0', 'e0', None,
       'e0', 'e0', None, None, 'e2', 'e2', None, 'e2', None, 'e2', None,
       'e2', 'e2', None, None, 'e2', 'e2', None, 'e2', None, 'e2', None,
       'e2', 'e2', None, None, 'e2', 'e2', None, 'e2', None, 'e2', None,
       'e2', 'e2', None, None]))

assert np.all(tt_no_repl.parent_name == np.array([
       None, 'e0', 'e0', None, 'e0', 'e0', None, None, 'e1', 'e1', None,
       'e1', 'e1', None, None, 'e2', 'e2', None, 'e2', None, 'e2', None,
       'e2', 'e2', None, None, 'e3', 'e3', None, 'e3', None, 'e3', None,
       'e3', 'e3', None, None, 'e4', 'e4', None, 'e4', None, 'e4', None,
       'e4', 'e4', None, None]))

assert_allclose(tt.s, [
       0.        , 0.        , 0.        , 0.5       , 0.5       ,
       1.        , 1.        , 1.        , 1.        , 1.        ,
       1.5       , 1.5       , 2.        , 2.        , 2.        ,
       2.        , 2.        , 2.16666667, 2.16666667, 2.5       ,
       2.5       , 2.83333333, 2.83333333, 3.        , 3.        ,
       3.        , 3.        , 3.        , 3.16666667, 3.16666667,
       3.5       , 3.5       , 3.83333333, 3.83333333, 4.        ,
       4.        , 4.        , 4.        , 4.        , 4.16666667,
       4.16666667, 4.5       , 4.5       , 4.83333333, 4.83333333,
       5.        , 5.        , 5.        ], rtol=0, atol=5e-9)

assert_allclose(tt_no_repl.s, [
       0.        , 0.        , 0.        , 0.5       , 0.5       ,
       1.        , 1.        , 1.        , 1.        , 1.        ,
       1.5       , 1.5       , 2.        , 2.        , 2.        ,
       2.        , 2.        , 2.16666667, 2.16666667, 2.5       ,
       2.5       , 2.83333333, 2.83333333, 3.        , 3.        ,
       3.        , 3.        , 3.        , 3.16666667, 3.16666667,
       3.5       , 3.5       , 3.83333333, 3.83333333, 4.        ,
       4.        , 4.        , 4.        , 4.        , 4.16666667,
       4.16666667, 4.5       , 4.5       , 4.83333333, 4.83333333,
       5.        , 5.        , 5.        ], rtol=0, atol=5e-9)

p1 = p0.copy()
p2 = p0.copy()

line.track(p1)
line_no_repl.track(p2)

assert_allclose(p2.x, p1.x, rtol=0, atol=1e-14)
assert_allclose(p2.px, p1.px, rtol=0, atol=1e-14)
assert_allclose(p2.y, p1.y, rtol=0, atol=1e-14)
assert_allclose(p2.py, p1.py, rtol=0, atol=1e-14)
assert_allclose(p2.zeta, p1.zeta, rtol=0, atol=1e-14)
assert_allclose(p2.delta, p1.delta, rtol=0, atol=1e-14)




from .environment import Environment
from .multiline_legacy import MultilineLegacy
import xtrack as xt

# For backward compatibility
class Multiline(Environment):

    def __init__(self, *args, **kwargs):
        print('Warning: Multiline is deprecated, use Environment instead')
        super().__init__(*args, **kwargs)

    @classmethod
    def from_dict(cls, dct, **kwargs):
        if 'xsuite_data_type' in dct and dct['xsuite_data_type'] == 'Environment':
            # TODO: Needs to be sorted, returns environment
            return super().from_dict(dct, **kwargs)
        else:
            out = cls._from_legacy_multiline_dict(dct)
            print('\nWarning: you seem to be loading a legacy multiline file. '
                  'The `Multiline` class is deprecated and is now replaced by `Environment`. '
                  'Your multiline has been converted automatically to an Environment object. '
                  '\nThis file will become unreadable in the future. We recommend to save it '
                  'as an Environment object. This can be easily done as follows:\n\n'
                  '    import xtrack as xt\n'
                  '    env = xt.Multiline.from_json("my_old_multiline.json")\n'
                  '    env.to_json("my_new_environment.json")\n')
            return out

    @classmethod
    def _from_legacy_multiline_dict(cls, dct):

        lines = {}

        for line_name in dct['lines']:

            dct_line = dct['lines'][line_name].copy()

            new_man_data = []
            for ee in dct['_var_manager']:
                new_ee = []
                skip = False
                for cc in ee:
                    if 'eref' in cc and f"eref['{line_name}']" not in cc:
                        skip = True
                        break
                    new_cc = cc.replace(f"eref['{line_name}']", 'element_refs')
                    new_ee.append(new_cc)

                if skip:
                    continue

                new_man_data.append(tuple(new_ee))

            dct_line['_var_management_data'] = {}
            dct_line['_var_management_data']['var_values'] = dct['_var_management_data']['var_values'].copy()
            dct_line['_var_manager'] = new_man_data

            line = xt.Line.from_dict(dct_line)

            lines[line_name] = line

        env = xt.Environment(lines=lines)
        if 'metadata' in dct:
            env.metadata.update(dct['metadata'])

        return env



import xtrack as xt
import numpy as np

env = xt.Environment()
env.particle_ref = xt.Particles(p0c=2e9)

n_bends_per_cell = 6
n_cells_par_arc = 3
n_arcs = 3

n_bends = n_bends_per_cell * n_cells_par_arc * n_arcs

env.vars({
    'l.mq': 0.5,
    'kqf': 0.027,
    'kqd': -0.0271,
    'l.mb': 10,
    'l.ms': 0.3,
    'k2sf': 0.001,
    'k2sd': -0.001,
    'angle.mb': 2 * np.pi / n_bends,
    'k0.mb': 'angle.mb / l.mb',
    'k0l.corrector': 0,
    'k1sl.corrector': 0,
    'l.halfcell': 38,
})

env.new('mb', xt.Bend, length='l.mb', k0='k0.mb', h='k0.mb')
env.new('mq', xt.Quadrupole, length='l.mq')
env.new('ms', xt.Sextupole, length='l.ms')
env.new('corrector', xt.Multipole, knl=[0], ksl=[0])

env.new('mq.f', 'mq', k1='kqf')
env.new('mq.d', 'mq', k1='kqd')

halfcell = env.new_line(components=[

    # End of the half cell (will be mid of the cell)
    env.new('mid', xt.Marker, at='l.halfcell'),

    # Bends
    env.new('mb.2', 'mb', at='l.halfcell / 2'),
    env.new('mb.1', 'mb', at='-l.mb - 1', from_='mb.2'),
    env.new('mb.3', 'mb', at='l.mb + 1', from_='mb.2'),

    # Quads
    env.place('mq.d', at = '0.5 + l.mq / 2'),
    env.place('mq.f', at = 'l.halfcell - l.mq / 2 - 0.5'),

    # Sextupoles
    env.new('ms.d', 'ms', k2='k2sf', at=1.2, from_='mq.d'),
    env.new('ms.f', 'ms', k2='k2sd', at=-1.2, from_='mq.f'),

    # Dipole correctors
    env.new('corrector.v', 'corrector', at=0.75, from_='mq.d'),
    env.new('corrector.h', 'corrector', at=-0.75, from_='mq.f')

])


cell = -halfcell + halfcell

opt = cell.match(
    method='4d',
    vary=xt.VaryList(['kqf', 'kqd'], step=1e-5),
    targets=xt.TargetSet(
        qx=0.333333,
        qy=0.333333,
    ))
tw_cell = cell.twiss4d()


env.vars({
    'kqf.ss': 0.027 / 2,
    'kqd.ss': -0.0271 / 2,
})

halfcell_ss = env.new_line(components=[

    env.new('mid', xt.Marker, at='l.halfcell'),

    env.new('mq.ss.d', 'mq', k1='kqd.ss', at = '0.5 + l.mq / 2'),
    env.new('mq.ss.f', 'mq', k1='kqf.ss', at = 'l.halfcell - l.mq / 2 - 0.5'),

    env.new('corrector.ss.v', 'corrector', at=0.75, from_='mq.ss.d'),
    env.new('corrector.ss.h', 'corrector', at=-0.75, from_='mq.ss.f')
])

cell_ss = env.new_line([
    env.new('start.cell.ss', 'Marker'),
    -halfcell_ss + halfcell_ss
])

opt = cell_ss.match(
    method='4d',
    vary=xt.VaryList(['kqf.ss', 'kqd.ss'], step=1e-5),
    targets=xt.TargetSet(
        betx=tw_cell.betx[-1], bety=tw_cell.bety[-1], at='start.cell.ss'))


arc = 3 * cell
ss = 2 * cell_ss

ring = 3 * (arc + ss)

# Twiss the ring
tw = ring.twiss4d()

# Twiss a single cell
tw_a_cell = ring.twiss4d(start='mq.f::10', end='mq.f::11', init='periodic')

## Insertion

env.vars({
    'k1.q1': 0.025,
    'k1.q2': -0.025,
    'k1.q3': 0.025,
    'k1.q4': -0.02,
    'k1.q5': 0.025,
})

half_insertion = env.new_line(components=[

    # Start-end markers
    env.new('ip', xt.Marker),
    env.new('e.insertion', xt.Marker, at=76),

    # Quads
    env.new('mq.1', xt.Quadrupole, k1='k1.q1', length='l.mq', at = 20),
    env.new('mq.2', xt.Quadrupole, k1='k1.q2', length='l.mq', at = 25),
    env.new('mq.3', xt.Quadrupole, k1='k1.q3', length='l.mq', at=37),
    env.new('mq.4', xt.Quadrupole, k1='k1.q4', length='l.mq', at=55),
    env.new('mq.5', xt.Quadrupole, k1='k1.q5', length='l.mq', at=73),

    # Dipole correctors (will use h and v on the same corrector)
    env.new('corrector.ss.1', 'corrector', at=0.75, from_='mq.1'),
    env.new('corrector.ss.2', 'corrector', at=-0.75, from_='mq.2'),
    env.new('corrector.ss.3', 'corrector', at=0.75, from_='mq.3'),
    env.new('corrector.ss.4', 'corrector', at=-0.75, from_='mq.4'),
    env.new('corrector.ss.5', 'corrector', at=0.75, from_='mq.5'),

])

tw_arc = arc.twiss4d()

opt = half_insertion.match(
    solve=False,
    betx=tw_arc.betx[0], bety=tw_arc.bety[0],
    alfx=tw_arc.alfx[0], alfy=tw_arc.alfy[0],
    init_at='e.insertion',
    start='ip', end='e.insertion',
    vary=xt.VaryList(['k1.q1', 'k1.q2', 'k1.q3', 'k1.q4'], step=1e-5),
    targets=[
        xt.TargetSet(alfx=0, alfy=0, at='ip'),
        xt.Target(lambda tw: tw.betx[0] - tw.bety[0], 0),
        xt.Target(lambda tw: tw.betx.max(), xt.LessThan(400)),
        xt.Target(lambda tw: tw.bety.max(), xt.LessThan(400)),
        xt.Target(lambda tw: tw.betx.min(), xt.GreaterThan(2)),
        xt.Target(lambda tw: tw.bety.min(), xt.GreaterThan(2)),
    ]
)
opt.step(40)
opt.solve()

insertion = -half_insertion + half_insertion

ring2 = 2 * (arc + ss) + arc + insertion

########################

## Line
ring.info("k0.mb") # Get info
ring.info('mb.3') # Get info

ring.get('k0.mb')# Get value of variable
ring.get('mb.3')# Get raw element
ring.get("k0.mb/2") # Error because it does not exists

ring.eval('k0.mb/2')# Evaluate string
ring.get_expr("k0.mb")# Get exprression associated to var
ring.new_expr("k0.mb/2")# Return new expression

## View
ring["mb.3"].info() ## print table
ring["mb.3"].info('k0') ## print info

ring["mb.3"].get_value('knl') # get the array
ring["mb.3"].get_expr('k0') ## get expresssion
ring["mb.3"].get('knl') # get the raw array  ##PROPOSAL

ring["mb.3"].get_value() # get a dictionary
ring["mb.3"].get_expr() # get dictionary of expressions
ring["mb.3"].to_dict() # get a dictionary ##PROPOSAL
ring["mb.3"].to_expr_dict() ## get a dictionary ##PROPOSAL


ring["mb.3"].get('knl') # get the array
ring["mb.3"].knl # get view the array

"""
Line     :        vars, ref, element_dict, ref_manager,  get_table
Env      : lines, vars, ref, element_dict, ref_manager
Env, Line: get, set, eval, get_expr, new_expr, info

vars     : get, set, eval, get_expr, new_expr, info, get_table
view     : get_expr, get_value, get_info, get_table
"""



import pytest
import xtrack as xt
import xobjects as xo
import pathlib
import numpy as np

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()

def test_madng_twiss():
    rdts = ["f4000", "f3100", "f2020", "f1120"]

    line = xt.Line.from_json(test_data_folder /
                            'hllhc15_thick/lhc_thick_with_knobs.json')

    line['test_dk1'] = 0
    line['mb.b32l8.b1'].knl[1] = 'test_dk1'

    tw = line.madng_twiss(normal_form=False)

    opt = line.match(
        solve=False,
        vary=[
            xt.Vary('on_x1', step=1e-3),
        ],
        targets=(
            tw.target('px_ng', 50e-6, at='ip1'),
            tw.target('py_ng', 0, at='ip1'),
        ),
    )
    opt.step(3)
    tw_after = line.madng_twiss(normal_form=False)
    xo.assert_allclose(tw_after['px_ng', 'ip1'], 50e-6, rtol=5e-3, atol=0)

    line['on_x1'] = 1.
    xo.assert_allclose(line.madng_twiss()['px_ng', 'ip1'], 1e-6, rtol=5e-3, atol=0)

    line['on_x1'] = -2.
    xo.assert_allclose(line.madng_twiss()['px_ng', 'ip1'], -2e-6, rtol=5e-3, atol=0)

    # We just check that rdts are there
    tw_rdt = line.madng_twiss(rdts=rdts)
    assert np.abs(tw_rdt.f4000).max() > 0
    assert np.abs(tw_rdt.f3100).max() > 0
    assert np.abs(tw_rdt.f2020).max() > 0
    assert np.abs(tw_rdt.f1120).max() > 0

def test_madng_interface_with_multipole_errors_and_misalignments():
    line = xt.Line.from_json(test_data_folder /
                            'hllhc15_thick/lhc_thick_with_knobs.json')

    tt = line.get_table()
    tt_quads = tt.rows[tt.element_type=='Quadrupole']

    # Introduce misalignments on all quadrupoles
    tt = line.get_table()
    tt_quad = tt.rows[r'mq\..*']
    rgen = np.random.RandomState(1) # fix seed for random number generator
                                    # (to have reproducible results)
    shift_x = rgen.randn(len(tt_quad)) * 0.01e-3 # 0.01 mm rms shift on all quads
    shift_y = rgen.randn(len(tt_quad)) * 0.01e-3 # 0.01 mm rms shift on all quads
    rot_s = rgen.randn(len(tt_quad)) * 1e-3 # 1 mrad rms rotation on all quads
    k2l = rgen.rand(len(tt_quad)) * 1e-3

    line['on_error'] = 1.
    for nn_quad, sx, sy, rr, kkk in zip(tt_quad.name, shift_x, shift_y, rot_s, k2l):
        line[nn_quad].shift_x = sx * line.ref['on_error']
        line[nn_quad].shift_y = sy * line.ref['on_error']
        line[nn_quad].rot_s_rad = rr * line.ref['on_error']
        line[nn_quad].knl[2] = kkk * line.ref['on_error']
    tw = line.madng_twiss()

    xo.assert_allclose(tw.x, tw.x_ng, atol=5e-4*tw.x.std(), rtol=0)
    xo.assert_allclose(tw.y, tw.y_ng, atol=5e-4*tw.y.std(), rtol=0)
    xo.assert_allclose(tw.betx2, tw.beta12_ng, atol=0, rtol=2e-3)
    xo.assert_allclose(tw.bety1, tw.beta21_ng, atol=0, rtol=2e-3)
    xo.assert_allclose(tw.wx_chrom, tw.wx_ng, atol=5e-3*tw.wx_chrom.max(), rtol=0)
    xo.assert_allclose(tw.wy_chrom, tw.wy_ng, atol=5e-3*tw.wy_chrom.max(), rtol=0)
    xo.assert_allclose(tw.ax_chrom, tw.ax_ng, atol=5e-3*tw.wx_chrom.max(), rtol=0)
    xo.assert_allclose(tw.ay_chrom, tw.ay_ng, atol=5e-3*tw.wy_chrom.max(), rtol=0)
    xo.assert_allclose(tw.bx_chrom, tw.bx_ng, atol=5e-3*tw.wx_chrom.max(), rtol=0)
    xo.assert_allclose(tw.by_chrom, tw.by_ng, atol=5e-3*tw.wy_chrom.max(), rtol=0)

    line['on_error'] = 0
    tw = line.madng_twiss()
    xo.assert_allclose(tw.x, 0, atol=1e-10, rtol=0)
    xo.assert_allclose(tw.y, 0, atol=1e-10, rtol=0)
    xo.assert_allclose(tw.betx2, 0, atol=1e-10, rtol=0)
    xo.assert_allclose(tw.bety1, 0, atol=1e-10, rtol=0)
    xo.assert_allclose(tw.x, tw.x_ng, atol=1e-10, rtol=0)
    xo.assert_allclose(tw.y, tw.y_ng, atol=1e-10, rtol=0)
    xo.assert_allclose(tw.betx2, tw.beta12_ng, atol=1e-10, rtol=0)
    xo.assert_allclose(tw.bety1, tw.beta21_ng, atol=1e-19, rtol=0)
    xo.assert_allclose(tw.wx_chrom, tw.wx_ng, atol=5e-3*tw.wx_chrom.max(), rtol=0)
    xo.assert_allclose(tw.wy_chrom, tw.wy_ng, atol=5e-3*tw.wy_chrom.max(), rtol=0)
    xo.assert_allclose(tw.ax_chrom, tw.ax_ng, atol=5e-3*tw.wx_chrom.max(), rtol=0)
    xo.assert_allclose(tw.ay_chrom, tw.ay_ng, atol=5e-3*tw.wy_chrom.max(), rtol=0)
    xo.assert_allclose(tw.bx_chrom, tw.bx_ng, atol=5e-3*tw.wx_chrom.max(), rtol=0)
    xo.assert_allclose(tw.by_chrom, tw.by_ng, atol=5e-3*tw.wy_chrom.max(), rtol=0)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from pathlib import Path
import numpy as np
from functools import partial

import xobjects as xo
from xobjects.general import Print

from xobjects.hybrid_class import _build_xofields_dict

from .general import _pkg_root
from .internal_record import RecordIdentifier, RecordIndex, generate_get_record
from .particles import Particles

start_per_part_block = """
    {
    const int64_t XT_part_block_start_idx = part0->ipart; //only_for_context cpu_openmp
    const int64_t XT_part_block_end_idx = part0->endpart; //only_for_context cpu_openmp

    const int64_t XT_part_block_start_idx = 0;                                            //only_for_context cpu_serial
    const int64_t XT_part_block_end_idx = LocalParticle_get__num_active_particles(part0); //only_for_context cpu_serial

    //#pragma omp simd // TODO: currently does not work, needs investigating
    for (int64_t XT_part_block_ii = XT_part_block_start_idx; XT_part_block_ii<XT_part_block_end_idx; XT_part_block_ii++) { //only_for_context cpu_openmp cpu_serial

        LocalParticle lpart = *part0;    //only_for_context cpu_serial cpu_openmp
        LocalParticle* part = &lpart;    //only_for_context cpu_serial cpu_openmp
        part->ipart = XT_part_block_ii;  //only_for_context cpu_serial cpu_openmp

        LocalParticle* part = part0;     //only_for_context opencl cuda

        if (LocalParticle_get_state(part) > 0) {  //only_for_context cpu_openmp
"""

end_part_part_block = """
        }  //only_for_context cpu_openmp
    }  //only_for_context cpu_serial cpu_openmp
    }
"""

def _handle_per_particle_blocks(sources, local_particle_src):

    if isinstance(sources, str):
        sources = (sources, )
        wasstring = True
    else:
        wasstring = False

    out = []
    for ii, ss in enumerate(sources):
        if isinstance(ss, Path):
            with open(ss, 'r') as fid:
                strss = fid.read()
        else:
            strss = ss

        strss = strss.replace('/*placeholder_for_local_particle_src*/',
                                local_particle_src,
                                )
        if '//start_per_particle_block' in strss:

            lines = strss.splitlines()
            for ill, ll in enumerate(lines):
                if '//start_per_particle_block' in ll:
                    lines[ill] = start_per_part_block
                if '//end_per_particle_block' in ll:
                    lines[ill] = end_part_part_block

            # TODO: this is very dirty, just for check!!!!!
            out.append('\n'.join(lines))
        else:
            out.append(ss)


    if wasstring:
        out = out[0]

    return out

def _generate_track_local_particle_with_transformations(
                                                element_name,
                                                allow_rot_and_shift,
                                                rot_and_shift_from_parent,
                                                local_particle_function_name):

    source = ('''
            /*gpufun*/
            '''
            f'void {local_particle_function_name}_with_transformations({element_name}Data el, LocalParticle* part0)'
            '{\n')

    if rot_and_shift_from_parent:
        add_to_call = '__parent'
    else:
        add_to_call = ''

    if allow_rot_and_shift:

        source += (
            '    // Transform to local frame\n'
            # f'    printf("Transform to local frame {element_name}\\n");\n'
            f'double const _sin_rot_s = {element_name}Data_get{add_to_call}__sin_rot_s(el);\n'
            'if (_sin_rot_s > -2.) {\n'
            f'    double const _cos_rot_s = {element_name}Data_get{add_to_call}__cos_rot_s(el);\n'
            f'    double const shift_x = {element_name}Data_get{add_to_call}__shift_x(el);\n'
            f'    double const shift_y = {element_name}Data_get{add_to_call}__shift_y(el);\n'
            f'    double const shift_s = {element_name}Data_get{add_to_call}__shift_s(el);\n'
            '\n'
            '    if (shift_s != 0.) {\n'
            '        //start_per_particle_block (part0->part)\n'
            '            Drift_single_particle(part, shift_s);\n'
            '        //end_per_particle_block\n'
            '    }\n'
            '\n'
            '    //start_per_particle_block (part0->part)\n'
            '       LocalParticle_add_to_x(part, -shift_x);\n'
            '       LocalParticle_add_to_y(part, -shift_y);\n'
            '    //end_per_particle_block\n'
            '\n'
            '     //start_per_particle_block (part0->part)\n'
            '          SRotation_single_particle(part, _sin_rot_s, _cos_rot_s);\n'
            '     //end_per_particle_block\n'
            '}\n'
        )

    source += (
            f'    {local_particle_function_name}(el, part0);\n'
    )

    if allow_rot_and_shift:
        source += (
            '    // Transform back to global frame\n'
            #f'    printf("Transform to back to global frame {element_name}\\n");\n'
            'if (_sin_rot_s > -2.) {\n'
            f'    double const _cos_rot_s = {element_name}Data_get{add_to_call}__cos_rot_s(el);\n'
            f'    double const shift_x = {element_name}Data_get{add_to_call}__shift_x(el);\n'
            f'    double const shift_y = {element_name}Data_get{add_to_call}__shift_y(el);\n'
            f'    double const shift_s = {element_name}Data_get{add_to_call}__shift_s(el);\n'
            '\n'
            '    //start_per_particle_block (part0->part)\n'
            '       SRotation_single_particle(part, -_sin_rot_s, _cos_rot_s);\n'
            '    //end_per_particle_block\n'
            '\n'
            '    //start_per_particle_block (part0->part)\n'
            '       LocalParticle_add_to_x(part, shift_x);\n'
            '       LocalParticle_add_to_y(part, shift_y);\n'
            '    //end_per_particle_block\n'
            '\n'
            '    if (shift_s != 0.) {\n'
            '        //start_per_particle_block (part0->part)\n'
            '            Drift_single_particle(part, -shift_s);\n'
            '        //end_per_particle_block\n'
            '    }\n'
            '}\n'
        )
    source += '}\n'
    return source

def _generate_per_particle_kernel_from_local_particle_function(
                                                element_name, kernel_name,
                                                local_particle_function_name,
                                                additional_args=[]):

    if len(additional_args) > 0:
        add_to_signature = ", ".join([
            f"{' /*gpuglmem*/ ' if arg.pointer else ''} {arg.get_c_type()} {arg.name}"
                for arg in additional_args]) + ", "
        add_to_call = ", " + ", ".join(f"{arg.name}" for arg in additional_args)

    source = ('''
            /*gpukern*/
            '''
            f'void {kernel_name}(\n'
            f'               {element_name}Data el,\n'
'''
                             ParticlesData particles,
'''
            f'{(add_to_signature if len(additional_args) > 0 else "")}'
'''
                             int64_t flag_increment_at_element,
                /*gpuglmem*/ int8_t* io_buffer){

            #define CONTEXT_OPENMP  //only_for_context cpu_openmp
            #ifdef CONTEXT_OPENMP
                const int64_t capacity = ParticlesData_get__capacity(particles);
                const int num_threads = omp_get_max_threads();

                #ifndef XT_OMP_SKIP_REORGANIZE
                    const int64_t num_particles_to_track = ParticlesData_get__num_active_particles(particles);

                    {
                        LocalParticle lpart;
                        lpart.io_buffer = io_buffer;
                        Particles_to_LocalParticle(particles, &lpart, 0, capacity);
                        check_is_active(&lpart);
                        count_reorganized_particles(&lpart);
                        LocalParticle_to_Particles(&lpart, particles, 0, capacity);
                    }
                #else // When we skip reorganize, we cannot just batch active particles
                    const int64_t num_particles_to_track = capacity;
                #endif

                const int64_t chunk_size = (num_particles_to_track + num_threads - 1)/num_threads; // ceil division
            #endif // CONTEXT_OPENMP

            #pragma omp parallel for                                                           //only_for_context cpu_openmp
            for (int64_t batch_id = 0; batch_id < num_threads; batch_id++) {                   //only_for_context cpu_openmp
                LocalParticle lpart;
                lpart.io_buffer = io_buffer;
                int64_t part_id = batch_id * chunk_size;                                       //only_for_context cpu_openmp
                int64_t end_id = (batch_id + 1) * chunk_size;                                  //only_for_context cpu_openmp
                if (end_id > num_particles_to_track) end_id = num_particles_to_track;          //only_for_context cpu_openmp

                int64_t part_id = 0;                    //only_for_context cpu_serial
                int64_t part_id = blockDim.x * blockIdx.x + threadIdx.x; //only_for_context cuda
                int64_t part_id = get_global_id(0);                    //only_for_context opencl
                int64_t end_id = 0; // unused outside of openmp  //only_for_context cpu_serial cuda opencl

                int64_t part_capacity = ParticlesData_get__capacity(particles);
                if (part_id<part_capacity){
                    Particles_to_LocalParticle(particles, &lpart, part_id, end_id);
                    if (check_is_active(&lpart)>0){
    '''
            f'          {local_particle_function_name}(el, &lpart{(add_to_call if len(additional_args) > 0 else "")});\n'
    '''
                    }
                    if (check_is_active(&lpart)>0 && flag_increment_at_element){
                            increment_at_element(&lpart, 1);
                    }
                }
            } //only_for_context cpu_openmp

            // On OpenMP we want to additionally by default reorganize all
            // the particles.
            #ifndef XT_OMP_SKIP_REORGANIZE                             //only_for_context cpu_openmp
            LocalParticle lpart;                                       //only_for_context cpu_openmp
            lpart.io_buffer = io_buffer;                               //only_for_context cpu_openmp
            Particles_to_LocalParticle(particles, &lpart, 0, capacity);//only_for_context cpu_openmp
            check_is_active(&lpart);                                   //only_for_context cpu_openmp
            #endif                                                     //only_for_context cpu_openmp
        }
''')
    return source

def _tranformations_active(self):

    if (self.shift_x == 0 and self.shift_y == 0 and self.shift_s == 0
        and self._sin_rot_s == 0 and self._cos_rot_s >= 0): # means angle is zero
        return False
    elif (self.shift_x == 0 and self.shift_y == 0 and self.shift_s == 0
          and self._sin_rot_s < -2.):
        return False
    else:
        return True

def _rot_s_property(self):
    if self._sin_rot_s < -2.:
        return 0.
    return np.arctan2(self._sin_rot_s, self._cos_rot_s)

def _set_rot_s_property_setter(self, value):
    self._sin_rot_s = np.sin(value)
    self._cos_rot_s = np.cos(value)
    if not _tranformations_active(self):
        self._sin_rot_s = -999.
    elif self._sin_rot_s < -2.:
        self._sin_rot_s = 0.
        self._cos_rot_s = 1.

def _shiftx_property(self):
    return self._shift_x

def _set_shiftx_property_setter(self, value):
    self._shift_x = value
    if not _tranformations_active(self):
        self._sin_rot_s = -999.
    elif self._sin_rot_s < -2.:
        self._sin_rot_s = 0.
        self._cos_rot_s = 1.

def _shifty_property(self):
    return self._shift_y

def _set_shifty_property_setter(self, value):
    self._shift_y = value
    if not _tranformations_active(self):
        self._sin_rot_s = -999.
    elif self._sin_rot_s < -2.:
        self._sin_rot_s = 0.
        self._cos_rot_s = 1.

def _shifts_property(self):
    return self._shift_s

def _set_shifts_property_setter(self, value):
    self._shift_s = value
    if not _tranformations_active(self):
        self._sin_rot_s = -999.
    elif self._sin_rot_s < -2.:
        self._sin_rot_s = 0.
        self._cos_rot_s = 1.

class MetaBeamElement(xo.MetaHybridClass):

    def __new__(cls, name, bases, data):
        _XoStruct_name = name+'Data'

        # Take xofields from data['_xofields'] or from bases
        xofields = _build_xofields_dict(bases, data)

        allow_rot_and_shift = data.get('allow_rot_and_shift', True)

        if allow_rot_and_shift:
            xofields['_sin_rot_s'] = xo.Field(xo.Float64, default=-999.)
            xofields['_cos_rot_s'] = xo.Field(xo.Float64, default=-999.)
            xofields['_shift_x'] = xo.Field(xo.Float64, 0)
            xofields['_shift_y'] = xo.Field(xo.Float64, 0)
            xofields['_shift_s'] = xo.Field(xo.Float64, 0)

        data = data.copy()
        data['_xofields'] = xofields

        depends_on = []
        extra_c_source = [
            _pkg_root.joinpath('headers','constants.h'),
            _pkg_root.joinpath('headers','checks.h'),
            _pkg_root.joinpath('headers','particle_states.h'),
            _pkg_root.joinpath('beam_elements', 'elements_src', 'track_srotation.h'),
            _pkg_root.joinpath('beam_elements', 'elements_src', 'drift.h'),
        ]
        kernels = {}

        # Handle internal record
        if '_internal_record_class' in data.keys():
            data['_xofields']['_internal_record_id'] = RecordIdentifier
            if '_skip_in_to_dict' not in data.keys():
                data['_skip_in_to_dict'] = []
            data['_skip_in_to_dict'].append('_internal_record_id')

            depends_on.append(RecordIndex)
            depends_on.append(data['_internal_record_class']._XoStruct)
            extra_c_source.append(
                generate_get_record(ele_classname=_XoStruct_name,
                    record_classname=data['_internal_record_class']._XoStruct.__name__))

        # Get user-defined source, dependencies and kernels
        if '_extra_c_sources' in data.keys():
            extra_c_source.extend(data['_extra_c_sources'])

        if '_depends_on' in data.keys():
            depends_on.extend(data['_depends_on'])

        if '_kernels' in data.keys():
            kernels.update(data['_kernels'])

        # Add dependency on Particles class
        depends_on.append(Particles._XoStruct)

        # For now I assume that when there is a parent, the element inherits the parent's transformations
        rot_and_shift_from_parent = False
        if '_parent' in xofields.keys():
            assert 'rot_and_shift_from_parent' in data.keys()
            rot_and_shift_from_parent = data['rot_and_shift_from_parent']

        track_kernel_name = None
        if ('allow_track' not in data.keys() or data['allow_track']):

            extra_c_source.append(
                _generate_track_local_particle_with_transformations(
                    element_name=name,
                    allow_rot_and_shift=(allow_rot_and_shift or rot_and_shift_from_parent),
                    rot_and_shift_from_parent=rot_and_shift_from_parent,
                    local_particle_function_name=name+'_track_local_particle'))

            # Generate track kernel
            extra_c_source.append(
                _generate_per_particle_kernel_from_local_particle_function(
                    element_name=name, kernel_name=name+'_track_particles',
                    local_particle_function_name=name+'_track_local_particle_with_transformations'))

            # Define track kernel
            track_kernel_name = f'{name}_track_particles'
            kernels[track_kernel_name] = xo.Kernel(
                c_name=track_kernel_name,
                args=[xo.Arg(xo.ThisClass, name='el'),
                    xo.Arg(Particles._XoStruct, name='particles'),
                    xo.Arg(xo.Int64, name='flag_increment_at_element'),
                    xo.Arg(xo.Int8, pointer=True, name="io_buffer")]
            )

        # Generate per-particle kernels
        if '_per_particle_kernels' in data.keys():
            for nn, kk in data['_per_particle_kernels'].items():
                extra_c_source.append(
                    _generate_per_particle_kernel_from_local_particle_function(
                        element_name=name, kernel_name=nn,
                        local_particle_function_name=kk.c_name,
                        additional_args=kk.args))
                if Particles._XoStruct not in depends_on:
                    depends_on.append(Particles._XoStruct)

                kernels.update(
                    {nn: xo.Kernel(
                        c_name=nn,
                        args=[
                        xo.Arg(xo.ThisClass, name='el'),
                        xo.Arg(Particles._XoStruct, name='particles'),
                        *kk.args,
                        xo.Arg(xo.Int64, name='flag_increment_at_element'),
                        xo.Arg(xo.Int8, pointer=True, name="io_buffer"),
                    ])}
                )

        # Call HybridClass metaclass
        data['_depends_on'] = depends_on
        data['_extra_c_sources'] = extra_c_source
        og_kernels = data.get('_kernels', {}).copy()
        data['_kernels'] = kernels
        new_class = xo.MetaHybridClass.__new__(cls, name, bases, data)

        # Attach some information to the class
        new_class._track_kernel_name = track_kernel_name
        if '_internal_record_class' in data.keys():
            new_class._XoStruct._internal_record_class = data['_internal_record_class']
            new_class._internal_record_class = data['_internal_record_class']

        # Attach methods corresponding to per-particle kernels
        if '_per_particle_kernels' in data.keys():
            for nn, desc in data['_per_particle_kernels'].items():
                setattr(new_class, nn, PerParticlePyMethodDescriptor(
                    kernel_name=nn,
                    additional_arg_names=tuple(arg.name for arg in desc.args),
                ))

        # Attach methods corresponding to kernels
        for nn, desc in og_kernels.items():
            setattr(new_class, nn, PyMethodDescriptor(
                kernel_name=nn,
                additional_arg_names=tuple(arg.name for arg in desc.args),
            ))

        if allow_rot_and_shift:
            new_class.rot_s_rad = property(_rot_s_property, _set_rot_s_property_setter)
            new_class.shift_x = property(_shiftx_property, _set_shiftx_property_setter)
            new_class.shift_y = property(_shifty_property, _set_shifty_property_setter)
            new_class.shift_s = property(_shifts_property, _set_shifts_property_setter)

        return new_class


class BeamElement(xo.HybridClass, metaclass=MetaBeamElement):

    iscollective = False
    isthick = False
    behaves_like_drift = False
    allow_track = True
    has_backtrack = False
    allow_loss_refinement = False
    allow_rot_and_shift = True
    skip_in_loss_location_refinement = False
    needs_rng = False
    name_associated_aperture = None

    def __init__(self, *args, **kwargs):
        xo.HybridClass.__init__(self, *args, **kwargs)

    def init_pipeline(self, pipeline_manager, name, partners_names=[]):
        self._pipeline_manager = pipeline_manager
        self.name = name
        self.partners_names = partners_names

    def compile_kernels(self, extra_classes=(), *args, **kwargs):
        if 'apply_to_source' not in kwargs.keys():
            kwargs['apply_to_source'] = []
        kwargs['apply_to_source'].append(
            partial(_handle_per_particle_blocks,
                    local_particle_src=Particles.gen_local_particle_api()))
        context = self._context
        cls = self.__class__

        if context.allow_prebuilt_kernels:
            # Default config is empty (all flags default to not defined, which
            # enables most behaviours). In the future this has to be looked at
            # whenever a new flag is needed.
            _default_config = {}
            _print_state = Print.suppress
            Print.suppress = True
            classes = (cls._XoStruct,) + tuple(extra_classes)
            try:
                from xsuite import (
                    get_suitable_kernel,
                    XSK_PREBUILT_KERNELS_LOCATION,
                )
            except ImportError:
                kernel_info = None
            else:
                kernel_info = get_suitable_kernel(
                    _default_config, classes
                )

            Print.suppress = _print_state
            if kernel_info:
                module_name, _ = kernel_info
                kernels = context.kernels_from_file(
                    module_name=module_name,
                    containing_dir=XSK_PREBUILT_KERNELS_LOCATION,
                    kernel_descriptions=self._kernels,
                )
                context.kernels.update(kernels)
                return
        xo.HybridClass.compile_kernels(
            self,
            extra_classes=[Particles._XoStruct],
            *args,
            **kwargs,
        )

    def track(self, particles=None, increment_at_element=False):
        if not self.allow_track:
            raise RuntimeError(f"BeamElement {self.__class__.__name__} "
                             + f"has no valid track method.")
        elif particles is None:
            raise RuntimeError("Please provide particles to track!")

        if self.needs_rng and not particles._has_valid_rng_state():
            particles._init_random_number_generator()

        context = self._buffer.context

        if self._track_kernel_name not in context.kernels:
            self.compile_kernels()

        _track_kernel = context.kernels[self._track_kernel_name]

        if hasattr(self, 'io_buffer') and self.io_buffer is not None:
            io_buffer_arr = self.io_buffer.buffer
        else:
            io_buffer_arr = context.zeros(1, dtype=np.int8)  # dummy

        _track_kernel.description.n_threads = particles._capacity
        _track_kernel(el=self._xobject, particles=particles,
                      flag_increment_at_element=increment_at_element,
                      io_buffer=io_buffer_arr)

    @property
    def context(self):
        return self._buffer.context

    def _arr2ctx(self, arr):
        ctx = self._buffer.context

        if isinstance(arr, list):
            arr = np.array(arr)

        if np.isscalar(arr):
            if hasattr(arr, 'item'):
                return arr.item()
            else:
                return arr
        elif isinstance(arr, ctx.nplike_array_type):
            return arr
        elif isinstance(arr, np.ndarray):
            return ctx.nparray_to_context_array(arr)
        else:
            raise ValueError("Invalid array type")

    def xoinitialize(self, **kwargs):
        rot_s_rad = kwargs.pop('rot_s_rad', None)
        shift_x = kwargs.pop('shift_x', None)
        shift_y = kwargs.pop('shift_y', None)
        shift_s = kwargs.pop('shift_s', None)

        xo.HybridClass.xoinitialize(self, **kwargs)

        if rot_s_rad is not None:
            self.rot_s_rad = rot_s_rad

        if shift_x is not None:
            self.shift_x = shift_x

        if shift_y is not None:
            self.shift_y = shift_y

        if shift_s is not None:
            self.shift_s = shift_s

    def to_dict(self, **kwargs):
        dct = xo.HybridClass.to_dict(self, **kwargs)
        if self.name_associated_aperture is not None:
            dct['name_associated_aperture'] = self.name_associated_aperture
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        if 'name_associated_aperture' in dct.keys():
            name_associated_aperture = dct.pop('name_associated_aperture')
        else:
            name_associated_aperture = None

        instance = xo.HybridClass._static_from_dict(cls, dct, **kwargs)
        instance.name_associated_aperture = name_associated_aperture
        return instance

    def copy(self, **kwargs):
        out = super().copy(**kwargs)
        if hasattr(self, 'extra'):
            try:
                out.extra = self.extra.copy()
            except AttributeError:
                out.extra = self.extra
        if hasattr(self, 'prototype'):
            out.prototype = self.prototype
        return out

    @property
    def _add_to_repr(self):
        out = []
        if hasattr(self, 'parent_name'):
            out.append('parent_name')
        return out

class Replica:
    def __init__(self, parent_name):
        self.parent_name = parent_name

    def __repr__(self):
        return f'Replica(parent_name="{self.parent_name}")'

    def to_dict(self):
        return {
            '__class__': 'Replica',
            'parent_name': self.parent_name}

    @classmethod
    def from_dict(cls, dct, **kwargs):
        return cls(parent_name=dct['parent_name'])

    def copy(self, **kwargs):
        return Replica(parent_name=self.parent_name)

    def resolve(self, element_container, get_name=False):
        if hasattr(element_container, 'element_dict'):
            element_container = element_container.element_dict
        target_name = self.parent_name
        visited = {target_name}
        while isinstance(element := element_container[target_name], Replica):
            target_name = element.parent_name
            if target_name in visited:
                raise RecursionError(
                    f"Resolving replica of `{self.parent_name}` leads to a "
                    "circular reference: check the correctness of your line."
                )
            visited.add(target_name)

        if get_name:
            return target_name

        return element_container[target_name]

class PerParticlePyMethod:

    def __init__(self, kernel_name, element, additional_arg_names):
        self.kernel_name = kernel_name
        self.element = element
        self.additional_arg_names = additional_arg_names

    def __call__(self, particles, increment_at_element=False, **kwargs):
        instance = self.element
        context = instance._context

        only_if_needed = kwargs.pop('only_if_needed', True)
        BeamElement.compile_kernels(instance, only_if_needed=only_if_needed)
        kernel = context.kernels[self.kernel_name]

        if hasattr(self.element, 'io_buffer') and self.element.io_buffer is not None:
            io_buffer_arr = self.element.io_buffer.buffer
        else:
            io_buffer_arr = context.zeros(1, dtype=np.int8)  # dummy

        kernel.description.n_threads = particles._capacity
        kernel(el=self.element._xobject,
               particles=particles,
               flag_increment_at_element=increment_at_element,
               io_buffer=io_buffer_arr,
               **kwargs)


class PerParticlePyMethodDescriptor:
    def __init__(self, kernel_name, additional_arg_names):
        self.kernel_name = kernel_name
        self.additional_arg_names = additional_arg_names

    def __get__(self, instance, owner):
        return PerParticlePyMethod(kernel_name=self.kernel_name,
                                   element=instance,
                                   additional_arg_names=self.additional_arg_names)


class PyMethod:

    def __init__(self, kernel_name, element, additional_arg_names):
        self.kernel_name = kernel_name
        self.element = element
        self.additional_arg_names = additional_arg_names

    def __call__(self, **kwargs):
        instance = self.element
        context = instance._context

        only_if_needed = kwargs.pop('only_if_needed', True)
        BeamElement.compile_kernels(instance, only_if_needed=only_if_needed)
        kernel = context.kernels[self.kernel_name]

        el_var_name = None
        for arg in instance._kernels[self.kernel_name].args:
            if arg.atype == instance.__class__._XoStruct:
                el_var_name = arg.name
        if el_var_name is None:
            raise ValueError(f"Kernel {self.kernel_name} does not depend "
                           + f"on element type {instance.__class__._XoStruct} "
                           + f"which it is attached to!")
        kwargs[el_var_name] = instance._xobject

        return kernel(**kwargs)


class PyMethodDescriptor:
    def __init__(self, kernel_name, additional_arg_names):
        self.kernel_name = kernel_name
        self.additional_arg_names = additional_arg_names

    def __get__(self, instance, owner):
        return PyMethod(kernel_name=self.kernel_name,
                        element=instance,
                        additional_arg_names=self.additional_arg_names)




from cpymad.madx import Madx
import xtrack as xt

# Things to check:
# - backtracking

mad = Madx()
mad.input("""


testseq: sequence, l=1;

squad1: quadrupole, k1s:=0.5, l=1, at=0.5;
! squad1: multipole, knl:={0, 0}, ksl:={0, 0.5}, at=1;

!quad: quadrupole, k1:=0.5, l=1, at=3;

!squad2: quadrupole, k1s:=-0.2, l=1, at=4;

endsequence;
beam;
use, sequence=testseq;
"""
)

line = xt.Line.from_madx_sequence(sequence=mad.sequence.testseq,
                                    deferred_expressions=True)
line.particle_ref = xt.Particles(p0c=7000e9, mass0=0.9382720813e9)

tw = line.twiss(x=1, betx=2, bety=1, start=line.element_names[0],
                                 end=line.element_names[-1])

twmad = mad.twiss(x=1, betx=2, bety=1, ripken=True)

print('betx2 mad:    ', twmad.beta12[-1])
print('betx2 xtrack: ', tw.betx2[-1])
print('bety1 mad:    ', twmad.beta21[-1])
print('bety1 xtrack: ', tw.bety1[-1])
print('x mad   :' , twmad.x[-1])
print('x xtrack:' , tw.x[-1])
print('y mad   :' , twmad.y[-1])
print('y xtrack:' , tw.y[-1])



class PipelineStatus:
    def __init__(self, on_hold, data=None, info=None):
        self.on_hold = on_hold
        self.data = data
        self.info = info

class PipelineID:
    def __init__(self,rank,number=0):
        self.number = number
        self.rank = rank

class PipelineCommunicator:

    def __init__(self):
        self.messages = {}

    def Issend(self,send_buffer,dest,tag):
        if dest not in self.messages.keys():
            self.messages[dest] = {}
        if tag not in self.messages[dest].keys():
            self.messages[dest][tag] = []
        self.messages[dest][tag].append(send_buffer)
        return self

    def Recv(self,receive_buffer,source,tag):
        assert source in self.messages.keys()
        assert tag in self.messages[source].keys()
        assert bool(self.messages[source][tag])
        message = self.messages[source][tag].pop(0)
        receive_buffer[:] = message[:]

    def Iprobe(self,source, tag):
        if source in self.messages.keys():
            if tag in self.messages[source].keys():
                return bool(self.messages[source][tag])
        return False

    def Test(self):
        return True



"""
Monitor to save the centroid of the many slices or time slots within a turn (like an actual BPM)

Author: Rahul Singh, Cristopher Cortes, Philipp Niedermayer
Date: 2023-06-10
"""

import numpy as np

import xobjects as xo

from ..base_element import BeamElement
from ..beam_elements import Marker
from ..internal_record import RecordIndex
from ..general import _pkg_root


class BeamPositionMonitorRecord(xo.Struct):
    count = xo.Float64[:]
    x_sum = xo.Float64[:]
    y_sum = xo.Float64[:]


class BeamPositionMonitor(BeamElement):

    _xofields={
        'particle_id_start': xo.Int64,
        'num_particles': xo.Int64,
        'start_at_turn': xo.Int64,
        'stop_at_turn': xo.Int64,
        'frev': xo.Float64,
        'sampling_frequency': xo.Float64,
        '_index': RecordIndex,
        'data': BeamPositionMonitorRecord,
    }

    behaves_like_drift = True
    allow_loss_refinement = True

    properties = [field.name for field in BeamPositionMonitorRecord._fields]

    _extra_c_sources = [
        _pkg_root.joinpath('headers/atomicadd.h'),
        _pkg_root.joinpath('monitors/beam_position_monitor.h')
    ]

    def __init__(self, *, particle_id_range=None, particle_id_start=None, num_particles=None,
                 start_at_turn=None, stop_at_turn=None, frev=None,
                 sampling_frequency=None, _xobject=None, **kwargs):
        """
        Monitor to save the transversal centroid of the tracked particles

        The monitor allows for arbitrary sampling rate and can thus not only be used to monitor
        bunch positions, but also to record schottky spectra. Internally, the particle arrival time
        is used when determining the record index:

            i = sampling_frequency * ( ( at_turn - start_turn ) / f_rev - zeta / beta0 / c0 )

        where zeta=(s-beta0*c0*t) is the longitudinal coordinate of the particle, beta0 the
        relativistic beta factor of the particle, c0 is the speed of light, at_turn is the
        current turn number, f_rev is the revolution frequency, and sampling_frequency is the
        sampling frequency.

        Note that the index is rounded, i.e. the result array represents data of particles
        equally distributed around the reference particle. For example, if the sampling_frequency
        is twice the revolution frequency, the first item contains data from particles in the
        range zeta/circumference = -0.25 .. 0.25, the second item in the range 0.25 .. 0.75 and
        so on.


        The monitor provides the following data:
        `count`, `x_sum`, `x_mean`, `y_sum`, `y_mean`,
        each as an array of size:
            size = int(( stop_at_turn - start_at_turn ) * sampling_frequency / frev)

        Args:
            num_particles (int, optional): Number of particles to monitor. Defaults to -1 which means ALL.
            particle_id_start (int, optional): First particle id to monitor. Defaults to 0.
            particle_id_range (tuple, optional): Range of particle ids to monitor (start, stop). Stop is exclusive.
                                                 Defaults to (particle_id_start, particle_id_start+num_particles).
            start_at_turn (int): First turn of reference particle (inclusive) at which to monitor.
            stop_at_turn (int): Last turn of reference particle (exclusiv) at which to monitor.
            frev (float): Revolution frequency in Hz of circulating beam (used to relate turn number to sample index).
            sampling_frequency (float): Sampling frequency in Hz.

        """
        if _xobject is not None:
            super().__init__(_xobject=_xobject)

        else:

            # dict parameters
            if particle_id_range is None:
                if particle_id_start is None:
                    particle_id_start = 0
                if num_particles is None:
                    num_particles = -1
            elif particle_id_start is None and num_particles is None:
                particle_id_start = particle_id_range[0]
                num_particles = particle_id_range[1] - particle_id_range[0]
            else:
                raise ValueError("Parameter `particle_id_range` must not be used together with `num_particles` and/or `particle_id_start`")
            if start_at_turn is None:
                start_at_turn = 0
            if stop_at_turn is None:
                stop_at_turn = 0
            if frev is None:
                frev = 1
            if sampling_frequency is None:
                sampling_frequency = 1
            if "data" not in kwargs:
                # explicitely init with zeros (instead of size only) to have consistent initial values
                size = int(round(( stop_at_turn - start_at_turn ) * sampling_frequency / frev))
                kwargs["data"] = {prop: np.zeros(size) for prop in self.properties}

            super().__init__(particle_id_start=particle_id_start, num_particles=num_particles,
                             start_at_turn=start_at_turn, stop_at_turn=stop_at_turn, frev=frev,
                             sampling_frequency=sampling_frequency, **kwargs)

    def __repr__(self):
        return (
            f"{type(self).__qualname__}(start_at_turn={self.start_at_turn}, stop_at_turn={self.stop_at_turn}, "
            f"particle_id_start={self.particle_id_start}, num_particles={self.num_particles}, frev={self.frev}, "
            f"sampling_frequency={self.sampling_frequency}) at {hex(id(self))}"
        )

    def __getattr__(self, attr):
        if attr in self.properties:
            return getattr(self.data, attr).to_nparray()

        if attr in ('x_mean', 'y_mean', 'x_cen', 'y_cen', 'x_centroid', 'y_centroid'):
            with np.errstate(invalid='ignore'):  # NaN for zero particles is expected behaviour
                return getattr(self, attr[0]+"_sum") / self.count

        return getattr(super(), attr)

    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return Marker(_context=_context, _buffer=_buffer, _offset=_offset)



import numpy as np
import xtrack as xt

def _compute_correction(x_iter, response_matrix, n_micado=None, rcond=None,
                        n_singular_values=None):

    if isinstance(response_matrix, (list, tuple)):
        assert len(response_matrix) == 3 # U, S, Vt
        U, S, Vh = response_matrix
        if n_singular_values is not None:
            U = U[:, :n_singular_values]
            S = S[:n_singular_values]
            Vh = Vh[:n_singular_values, :]
        response_matrix = U @ np.diag(S) @ Vh
    else:
        assert n_singular_values is None
        U = None
        S = None
        Vh = None

    n_hcorrectors = response_matrix.shape[1]

    if n_micado is not None:
        used_correctors = []

        for i_micado in range(n_micado):

            residuals = []
            for i_corr in range(n_hcorrectors):
                if i_corr in used_correctors:
                    residuals.append(np.nan)
                    continue
                mask_corr = np.zeros(n_hcorrectors, dtype=bool)
                mask_corr[i_corr] = True
                for i_used in used_correctors:
                    mask_corr[i_used] = True

                # Compute the correction with least squares
                _, residual_x, rank_x, sval_x = np.linalg.lstsq(
                            response_matrix[:, mask_corr], -x_iter, rcond=rcond)
                residuals.append(residual_x[0])
            used_correctors.append(np.nanargmin(residuals))

        mask_corr = np.zeros(n_hcorrectors, dtype=bool)
        mask_corr[used_correctors] = True
    else:
        mask_corr = np.ones(n_hcorrectors, dtype=bool)
        mask_corr[:] = True

    # Compute the correction with least squares
    if mask_corr.all() and S is not None:
        # Can reuse the SVD decomposition
        S_inv = np.zeros_like(S)
        S_inv[S > 0] = 1 / S[S > 0]
        if rcond is not None:
            S_inv[S < rcond * S[0]] = 0
        correction_masked = Vh.T.conj() @ (np.diag(S_inv) @ (U.T.conj() @ (-x_iter)))
    else:
        correction_masked, residual_x, rank_x, sval_x = np.linalg.lstsq(
                    response_matrix[:, mask_corr], -x_iter, rcond=rcond)
    correction_x = np.zeros(n_hcorrectors)
    correction_x[mask_corr] = correction_masked

    return correction_x


def _build_response_matrix(tw, monitor_names, corrector_names,
                           mode='closed', plane=None):

    assert mode in ['closed', 'open']
    assert plane in ['x', 'y']

    # Build response matrix
    indices_monitors = tw.rows.indices[monitor_names]
    indices_correctors = tw.rows.indices[corrector_names]
    bet_monitors = tw['bet' + plane][indices_monitors]
    bet_correctors = tw['bet' + plane][indices_correctors]

    mu_monitor = tw['mu' + plane][indices_monitors]
    mux_correctors = tw['mu' + plane][indices_correctors]

    n_monitors = len(monitor_names)
    n_correctors = len(corrector_names)

    bet_prod = np.atleast_2d(bet_monitors).T @ np.atleast_2d(bet_correctors)
    mu_diff = (np.tile(mu_monitor, (n_correctors, 1)).T
                        - np.tile(mux_correctors, (n_monitors, 1)))

    if mode == 'open':
        # Wille eq. 3.164
        mu_diff[mu_diff < 0] = 0 # use only correctors upstream of the monitor
        response_matrix = (np.sqrt(bet_prod) * np.sin(2*np.pi*np.abs(mu_diff)))
    elif mode == 'closed':
        # Slide 28
        # https://indico.cern.ch/event/1328128/contributions/5589794/attachments/2786478/4858384/linearimperfections_2024.pdf
        tune = tw.qx
        response_matrix = (np.sqrt(bet_prod) / 2 / np.sin(np.pi * tune)
                             * np.cos(np.pi * tune - 2*np.pi*np.abs(mu_diff)))

    return response_matrix


class OrbitCorrectionSinglePlane:

    def __init__(self, line, plane, monitor_names, corrector_names,
                 start=None, end=None, twiss_table=None, n_micado=None,
                 n_singular_values=None, rcond=None,
                 x_init=0, px_init=0, y_init=0, py_init=0, zeta_init=0, delta_init=0,
                 monitor_alignment=None):

        assert plane in ['x', 'y']

        self.x_init = x_init
        self.px_init = px_init
        self.y_init = y_init
        self.py_init = py_init
        self.zeta_init = zeta_init
        self.delta_init = delta_init

        self.twiss_table = twiss_table
        if twiss_table is not None:
            assert twiss_table.reference_frame == 'proper', (
                'Twiss table must be in the proper frame (`reverse` not supported)')

        if start is None:
            assert end is None
            self.mode = 'closed'
            if self.twiss_table is None:
                self.twiss_table = line.twiss4d(reverse=False)
        else:
            assert end is not None
            self.mode = 'open'
            if self.twiss_table is None:
                # Initialized with betx=1, bety=1 (use W_matrix to avoid compilation)
                self.twiss_table = line.twiss4d(start=start, end=end,
                    init=xt.TwissInit(W_matrix=np.eye(6),
                                      particle_on_co=line.build_particles(
                                          x=self.x_init, y=self.y_init,
                                          px=self.px_init, py=self.py_init,
                                          zeta=self.zeta_init, delta=self.delta_init),
                                      element_name=start),
                                      reverse=False)

        if corrector_names is None:
            corr_names_from_line = getattr(line, f'steering_correctors_{plane}')
            assert corr_names_from_line is not None, (
                f'No steering correctors found for plane {plane}')
            if start is not None:
                corrector_names = [nn for nn in corr_names_from_line
                                   if nn in self.twiss_table.name]
            else:
                corrector_names = corr_names_from_line

        if monitor_names is None:
            monitor_names = getattr(line, f'steering_monitors_{plane}')
            assert monitor_names is not None, (
                f'No monitors found for plane {plane}')
            if start is not None:
                monitor_names = [nn for nn in monitor_names
                                 if nn in self.twiss_table.name]
            else:
                monitor_names = monitor_names

        assert len(monitor_names) > 0
        assert len(corrector_names) > 0

        self.line = line
        self.plane = plane
        self.monitor_names = monitor_names
        self.corrector_names = corrector_names
        self.start = start
        self.end = end
        self.n_micado = n_micado
        self.rcond = rcond
        self.n_singular_values = n_singular_values

        self.response_matrix = _build_response_matrix(plane=self.plane,
            tw=self.twiss_table, monitor_names=self.monitor_names,
            corrector_names=self.corrector_names, mode=self.mode)

        U, S , Vt = np.linalg.svd(self.response_matrix, full_matrices=False)
        self.singular_values = S
        self.singular_vectors_out = U
        self.singular_vectors_in = Vt

        tw_table_local = self.twiss_table.rows[start:end]

        self._indices_monitor = tw_table_local.rows.indices[self.monitor_names]
        self._indices_correctors = tw_table_local.rows.indices[self.corrector_names]
        self.s_correctors = tw_table_local.s[self._indices_correctors]
        self.s_monitors = tw_table_local.s[self._indices_monitor]

        self.shift_x_monitors = 0*self.s_monitors
        self.shift_y_monitors = 0*self.s_monitors
        self.rot_s_rad_monitors = 0*self.s_monitors

        self.monitor_alignment = monitor_alignment
        if monitor_alignment is not None:
            for nn, alignment in monitor_alignment.items():
                for kk in alignment.keys():
                    assert kk in ['shift_x', 'shift_y', 'rot_s_rad']
                if nn in self.monitor_names:
                    i_monitor = self.monitor_names.index(nn)
                    self.shift_x_monitors[i_monitor] = alignment.get('shift_x', 0)
                    self.shift_y_monitors[i_monitor] = alignment.get('shift_y', 0)
                    self.rot_s_rad_monitors[i_monitor] = alignment.get('rot_s_rad', 0)

        self._add_correction_knobs()

    def correct(self, n_iter='auto', n_micado=None, n_singular_values=None,
                rcond=None, stop_iter_factor=0.1, verbose=True, _tw_orbit=None):

        if _tw_orbit is not None and n_iter !=1:
            raise ValueError('`_tw_orbit` can only be used with `n_iter=1`')

        assert n_iter == 'auto' or np.isscalar(n_iter)
        if n_iter == 'auto':
            assert stop_iter_factor > 0
            assert stop_iter_factor < 1

        pos_rms_prev = 0

        i_iter = 0
        while True:
            try:
                position = self._measure_position(tw_orbit=_tw_orbit)
            except xt.twiss.ClosedOrbitSearchError:
                raise RuntimeError('Closed orbit not found. '
                    'Please use the `thread(...)` method to obtain a first guess, '
                    'then call the `correct(...)` method again.')
            self._position_before_iter = position
            if i_iter == 0:
                self._position_before = position
            if verbose:
                print(
                    f'Trajectory correction - iter {i_iter}, rms: {position.std()}')

            if n_iter == 'auto':
                if i_iter > 0 and position.std() > (1. - stop_iter_factor) * pos_rms_prev:
                    break
                pos_rms_prev = position.std()

            correction = self._compute_correction(position=position,
                                    n_micado=n_micado, rcond=rcond,
                                    n_singular_values=n_singular_values)
            self._apply_correction(correction)

            i_iter += 1
            if n_iter != 'auto' and i_iter >= n_iter:
                break

        if _tw_orbit is None:
            position = self._measure_position()
            self._position_after = position
            if verbose:
                print(
                    f'Trajectory correction - iter {i_iter}, rms: {position.std()}')
        else:
            self._position_after = None

    def _compute_tw_orbit(self):
        if self.mode == 'open':
            # Initialized with betx=1, bety=1 (use W_matrix to avoid compilation)
            twinit = xt.TwissInit(W_matrix=np.eye(6),
                            particle_on_co=self.line.build_particles(
                                          x=self.x_init, y=self.y_init,
                                            px=self.px_init, py=self.py_init,
                                            zeta=self.zeta_init, delta=self.delta_init),
                            element_name=self.start)
        else:
            twinit = None
        tw_orbit = self.line.twiss4d(only_orbit=True, start=self.start, end=self.end,
                                     init=twinit, reverse=False)
        return tw_orbit

    def _measure_position(self, tw_orbit=None):

        if tw_orbit is None:
            tw_orbit = self._compute_tw_orbit()

        x_twiss_at_monitor = tw_orbit['x'][self._indices_monitor]
        y_twiss_at_monitor = tw_orbit['y'][self._indices_monitor]

        x_shifted = x_twiss_at_monitor - self.shift_x_monitors
        y_shifted = y_twiss_at_monitor - self.shift_y_monitors

        x_local = x_shifted * np.cos(self.rot_s_rad_monitors) + y_shifted * np.sin(self.rot_s_rad_monitors)
        y_local = -x_shifted * np.sin(self.rot_s_rad_monitors) + y_shifted * np.cos(self.rot_s_rad_monitors)

        position = {'x': x_local, 'y': y_local}[self.plane]

        return position

    def _compute_correction(self, position, n_micado=None,
                            n_singular_values=None, rcond=None):

        if rcond is None:
            rcond = self.rcond

        if n_singular_values is None:
            n_singular_values = self.n_singular_values

        if n_micado is None:
            n_micado = self.n_micado

        correction = _compute_correction(position,
            response_matrix=(self.singular_vectors_out, self.singular_values, self.singular_vectors_in),
            n_micado=n_micado,
            rcond=rcond, n_singular_values=n_singular_values)

        return correction

    def _add_correction_knobs(self):

        self.correction_knobs = []
        for nn_kick in self.corrector_names:
            corr_knob_name = f'orbit_corr_{nn_kick}_{self.plane}'
            assert hasattr(self.line[nn_kick], 'knl')
            assert hasattr(self.line[nn_kick], 'ksl')

            if corr_knob_name not in self.line.vars:
                self.line.vars[corr_knob_name] = 0

            if self.plane == 'x':
                if (self.line.element_refs[nn_kick].knl[0]._expr is None or
                    (self.line.vars[corr_knob_name]
                    not in self.line.element_refs[nn_kick].knl[0]._expr._get_dependencies())):
                    if self.line.element_refs[nn_kick].knl[0]._expr is not None:
                        self.line.element_refs[nn_kick].knl[0] -= ( # knl[0] is -kick
                            self.line.vars[f'orbit_corr_{nn_kick}_x'])
                    else:
                        # Workarond for https://github.com/xsuite/xsuite/issues/501
                        val = self.line.element_refs[nn_kick].knl[0]._value
                        if hasattr(val, 'get'):
                            val = val.get()
                        self.line.element_refs[nn_kick].knl[0] = val - (
                            self.line.vars[f'orbit_corr_{nn_kick}_x'])
            elif self.plane == 'y':
                if (self.line.element_refs[nn_kick].ksl[0]._expr is None or
                    (self.line.vars[corr_knob_name]
                    not in self.line.element_refs[nn_kick].ksl[0]._expr._get_dependencies())):
                    if  self.line.element_refs[nn_kick].ksl[0]._expr is not None:
                        self.line.element_refs[nn_kick].ksl[0] += ( # ksl[0] is +kick
                            self.line.vars[f'orbit_corr_{nn_kick}_y'])
                    else:
                        # Workarond for https://github.com/xsuite/xsuite/issues/501
                        val = self.line.element_refs[nn_kick].ksl[0]._value
                        if hasattr(val, 'get'):
                            val = val.get()
                        self.line.element_refs[nn_kick].ksl[0] = val + (
                            self.line.vars[f'orbit_corr_{nn_kick}_y'])

            self.correction_knobs.append(corr_knob_name)

    def _apply_correction(self, correction=None):

        if correction is None:
            correction = self.correction

        for nn_knob, kick in zip(self.correction_knobs, correction):
            self.line.vars[nn_knob] += kick

    def get_kick_values(self):
        return np.array([self.line.vv[nn_knob] for nn_knob in self.correction_knobs])

    def clear_correction_knobs(self):
        for nn_knob in self.correction_knobs:
            self.line.vars[nn_knob] = 0

class TrajectoryCorrection:

    def __init__(self, line,
                 start=None, end=None, twiss_table=None,
                 monitor_names_x=None, corrector_names_x=None,
                 monitor_names_y=None, corrector_names_y=None,
                 monitor_alignment=None,
                 x_init=0, px_init=0, y_init=0, py_init=0, zeta_init=0, delta_init=0,
                 n_micado=None, n_singular_values=None, rcond=None):

        '''
        Trajectory correction using linearized response matrix from optics
        table.

        Parameters
        ----------

        line : xtrack.Line
            Line object on which the trajectory correction is performed.
                start : str
            Start of the line range in which the correction is performed.
            If `start` is provided `end` must also be provided.
            If `start` is None, the correction is performed on the periodic
            solution (closed orbit).
        end : str
            End of the line range in which the correction is performed.
            If `end` is provided `start` must also be provided.
            If `start` is None, the correction is performed on the periodic
            solution (closed orbit).
        twiss_table : TwissTable
            Twiss table used to compute the response matrix for the correction.
            If None, the twiss table is computed from the line.
        monitor_names_x : list of str
            List of elements used as monitors in the horizontal plane.
        corrector_names_x : list of str
            List of elements used as correctors in the horizontal plane. They
            must have `knl` and `ksl` attributes.
        monitor_names_y : list of str
            List of elements used as monitors in the vertical plane.
        corrector_names_y : list of str
            List of elements used as correctors in the vertical plane. They
            must have `knl` and `ksl` attributes.
        n_micado : int
            If `n_micado` is not None, the MICADO algorithm is used for the
            correction. In that case, the number of correctors to be used is
            given by `n_micado`.
        n_singular_values : int
            Number of singular values used for the correction.
        rcond : float
            Cutoff for small singular values (relative to the largest singular
            value). Singular values smaller than `rcond` are considered zero.
        '''

        if isinstance(rcond, (tuple, list)):
            rcond_x, rcond_y = rcond
        else:
            rcond_x, rcond_y = rcond, rcond

        if isinstance(n_singular_values, (tuple, list)):
            n_singular_values_x, n_singular_values_y = n_singular_values
        else:
            n_singular_values_x, n_singular_values_y = n_singular_values, n_singular_values

        if isinstance(n_micado, (tuple, list)):
            n_micado_x, n_micado_y = n_micado
        else:
            n_micado_x, n_micado_y = n_micado, n_micado

        self.x_init = x_init
        self.px_init = px_init
        self.y_init = y_init
        self.py_init = py_init
        self.zeta_init = zeta_init
        self.delta_init = delta_init

        self.monitor_alignment = monitor_alignment

        if (monitor_names_x is not None or corrector_names_x is not None
            or line.steering_correctors_x is not None
            or line.steering_monitors_x is not None):
            self.x_correction = OrbitCorrectionSinglePlane(
                line=line, plane='x', monitor_names=monitor_names_x,
                corrector_names=corrector_names_x, start=start, end=end,
                twiss_table=twiss_table, n_micado=n_micado_x,
                n_singular_values=n_singular_values_x, rcond=rcond_x,
                x_init=x_init, px_init=px_init, y_init=y_init, py_init=py_init,
                zeta_init=zeta_init, delta_init=delta_init,
                monitor_alignment=monitor_alignment)
        else:
            self.x_correction = None

        if (monitor_names_y is not None or corrector_names_y is not None
           or line.steering_correctors_y is not None
           or line.steering_monitors_y is not None):
            self.y_correction = OrbitCorrectionSinglePlane(
                line=line, plane='y', monitor_names=monitor_names_y,
                corrector_names=corrector_names_y, start=start, end=end,
                twiss_table=twiss_table, n_micado=n_micado_y,
                n_singular_values=n_singular_values_y, rcond=rcond_y,
                x_init=x_init, px_init=px_init, y_init=y_init, py_init=py_init,
                zeta_init=zeta_init, delta_init=delta_init,
                monitor_alignment=monitor_alignment)
        else:
            self.y_correction = None

    def correct(self, planes=None, n_micado=None, n_singular_values=None,
                rcond=None, n_iter='auto', verbose=True, stop_iter_factor=0.1,
                tol_position_std=1e-10):

        '''
        Correct the trajectory in the horizontal and/or vertical plane.

        Parameters
        ----------
        planes : str
            Plane(s) in which the correction is performed. Possible values are
            'x', 'y', 'xy'.
        n_micado : int or tuple of int
            If `n_micado` is not None, the MICADO algorithm is used for the
            correction. In that case, the number of correctors to be used is
            given by `n_micado`.
        n_singular_values : int or tuple of int
            Number of singular values used for the correction.
        rcond : float or tuple of float
            Cutoff for small singular values (relative to the largest singular
            value). Singular values smaller than `rcond` are considered zero.
        n_iter : int or 'auto'
            Number of iterations for the correction. If 'auto', the correction
            stops when the rms of the position does not decrease by more than
            `stop_iter_factor` with respect to the previous iteration.
        verbose : bool
            If True, print the rms of the position at each iteration.
        stop_iter_factor : float
            If `n_iter` is 'auto', the correction stops when the rms of the
            position does not decrease by more than `stop_iter_factor` with
            respect to the previous iteration.
        '''

        assert n_iter == 'auto' or np.isscalar(n_iter)
        if n_iter == 'auto':
            assert stop_iter_factor > 0
            assert stop_iter_factor < 1

        if isinstance(rcond, (tuple, list)):
            rcond_x, rcond_y = rcond
        else:
            rcond_x, rcond_y = rcond, rcond

        if isinstance(n_singular_values, (tuple, list)):
            n_singular_values_x, n_singular_values_y = n_singular_values
        else:
            n_singular_values_x, n_singular_values_y = n_singular_values, n_singular_values

        if isinstance(n_micado, (tuple, list)):
            n_micado_x, n_micado_y = n_micado
        else:
            n_micado_x, n_micado_y = n_micado, n_micado

        if planes is None:
            planes = 'xy'
        assert planes in ['x', 'y', 'xy']

        i_iter = 0
        stop_x = self.x_correction is None or 'x' not in planes
        stop_y = self.y_correction is None or 'y' not in planes

        if stop_x and stop_y:
            return

        if self.x_correction is not None:
            a_correction = self.x_correction
        if self.y_correction is not None:
            a_correction = self.y_correction

        tw_orbit = a_correction._compute_tw_orbit()

        while True:

            if self.x_correction is not None and 'x' in planes:
                self.x_correction.correct(n_micado=n_micado_x,
                            n_singular_values=n_singular_values_x,
                            rcond=rcond_x, verbose=False, n_iter=1,
                            _tw_orbit=tw_orbit)

            if self.y_correction is not None and 'y' in planes:
                self.y_correction.correct(n_micado=n_micado_y,
                            n_singular_values=n_singular_values_y,
                            rcond=rcond_y, verbose=False, n_iter=1,
                            _tw_orbit=tw_orbit)

            tw_orbit_prev = tw_orbit
            tw_orbit = a_correction._compute_tw_orbit()

            if n_iter == 'auto' and self.x_correction is not None and 'x' in planes:
                new_position = self.x_correction._measure_position(tw_orbit)
                old_position = self.x_correction._measure_position(tw_orbit_prev)
                stop_x = (new_position.std() < tol_position_std or
                    new_position.std() > (1. - stop_iter_factor) * old_position.std())

            if n_iter == 'auto' and self.y_correction is not None and 'y' in planes:
                new_position = self.y_correction._measure_position(tw_orbit)
                old_position = self.y_correction._measure_position(tw_orbit_prev)
                stop_y = (new_position.std() < tol_position_std or
                    new_position.std() > (1. - stop_iter_factor) * old_position.std())

            if verbose:
                str_2print = f'Iteration {i_iter}, '
                if self.x_correction is not None and 'x' in planes:
                    new_position = self.x_correction._measure_position(tw_orbit)
                    old_position = self.x_correction._measure_position(tw_orbit_prev)
                    str_2print += (f'x_rms: {old_position.std():.2e}'
                        f' -> {new_position.std():.2e}, ')
                if self.y_correction is not None and 'y' in planes:
                    new_position = self.y_correction._measure_position(tw_orbit)
                    old_position = self.y_correction._measure_position(tw_orbit_prev)
                    str_2print += (f'y_rms: {old_position.std():.2e}'
                        f' -> {new_position.std():.2e}')
                print(str_2print)
            if stop_x and stop_y:
                break
            i_iter += 1
            if n_iter != 'auto' and i_iter >= n_iter:
                break

    def thread(self, ds_thread=None, rcond_short=None, rcond_long=None):

        '''
        Thread the trajectory along the line. The correction is performed in
        portions of length `ds_thread`. For each portion the correction is
        firs performed only on the new added part, then on the whole portion up
        to the end of the new added part.

        Parameters
        ----------
        ds_thread : float
            Length of the portion added at each iteration.
        rcond_long : float or tuple of float
            Cutoff for small singular values (relative to the largest singular
            value) used for the correction of the whole portion up to the end
            of the new added part.
        '''

        if self.start is not None or self.end is not None:
            raise NotImplementedError('Thread not implemented for line portions')

        threader = _thread(line=self.line, ds_thread=ds_thread, twiss_table=self.twiss_table,
                rcond_short=rcond_short, rcond_long=rcond_long,
                monitor_names_x=self.x_correction.monitor_names,
                monitor_names_y=self.y_correction.monitor_names,
                corrector_names_x=self.x_correction.corrector_names,
                corrector_names_y=self.y_correction.corrector_names,
                monitor_alignment=self.monitor_alignment)
        return threader

    def clear_correction_knobs(self):

        '''
        Set all correction knobs to zero. Erases all applied corrections.
        '''

        if self.x_correction is not None:
            self.x_correction.clear_correction_knobs()
        if self.y_correction is not None:
            self.y_correction.clear_correction_knobs()

    @property
    def start(self):
        if self.x_correction is not None:
            x_start = self.x_correction.start
        else:
            x_start = None
        if self.y_correction is not None:
            y_start = self.y_correction.start
        else:
            y_start = None

        if x_start is None:
            return y_start
        if y_start is None:
            return x_start
        if x_start is not None and y_start is not None:
            assert x_start == y_start
            return x_start

    @property
    def end(self):
        if self.x_correction is not None:
            x_end = self.x_correction.end
        else:
            x_end = None
        if self.y_correction is not None:
            y_end = self.y_correction.end
        else:
            y_end = None

        if x_end is None:
            return y_end
        if y_end is None:
            return x_end
        if x_end is not None and y_end is not None:
            assert x_end == y_end
            return x_end
    @property
    def line(self):
        if self.x_correction is not None:
            return self.x_correction.line
        if self.y_correction is not None:
            return self.y_correction.line

    @property
    def twiss_table(self):
        if self.x_correction is not None:
            return self.x_correction.twiss_table
        if self.y_correction is not None:
            return self.y_correction.twiss_table

def _thread(line, ds_thread, twiss_table=None, rcond_short = None, rcond_long = None,
            monitor_names_x=None, monitor_names_y=None,
            corrector_names_x=None, corrector_names_y=None,
            x_init=0, px_init=0, y_init=0, py_init=0, zeta_init=0, delta_init=0,
            monitor_alignment=None,
            verbose=True):

    assert ds_thread is not None

    # r_cond_short is not used anymore, see commented code below

    tt = line.get_table()
    line_length = tt.s[-1]

    if monitor_names_x is None:
        monitor_names_x = line.steering_monitors_x

    if monitor_names_y is None:
        monitor_names_y = line.steering_monitors_y

    if corrector_names_x is None:
        corrector_names_x = line.steering_correctors_x

    if corrector_names_y is None:
        corrector_names_y = line.steering_correctors_y

    i_win = 0
    end_loop = False
    s_corr_end = ds_thread
    while not end_loop:

        if s_corr_end > line_length:
            s_corr_end = line_length
            end_loop = True

        # ----- The following was used to correct only the newly added part
        # ----- It us not used anymore, as it was observed not to help
        # ----- We keep it in case it is needed in the futures
        # Correct only the new added portion
        # tt_new_part = tt.rows[s_corr_end-ds_thread:s_corr_end:'s']
        #
        # Get initial conditions for the new added portion
        # if i_win == 0:
        #     this_x_init = x_init
        #     this_y_init = y_init
        #     this_px_init = px_init
        #     this_py_init = py_init
        #     this_zeta_init = zeta_init
        #     this_delta_init = delta_init
        # else:
        #     # Initialized with betx=1, bety=1 (use W_matrix to avoid compilation)
        #     name_start = tt_new_part.name[0]
        #     tw_to_start = line.twiss4d(
        #         start=line.element_names[0], end=tt_new_part.name[0],
        #             init=xt.TwissInit(W_matrix=np.eye(6),
        #                                 particle_on_co=line.build_particles(
        #                                     x=x_init, y=y_init,
        #                                     px=px_init, py=py_init,
        #                                     zeta=zeta_init, delta=delta_init),
        #                                 element_name=line.element_names[0]),
        #                                 reverse=False)

        #     this_x_init = tw_to_start['x', name_start]
        #     this_y_init = tw_to_start['y', name_start]
        #     this_px_init = tw_to_start['px', name_start]
        #     this_py_init = tw_to_start['py', name_start]
        #     this_zeta_init = tw_to_start['zeta', name_start]
        #     this_delta_init = tw_to_start['delta', name_start]

        # ocorr_only_added_part = TrajectoryCorrection(
        #     line=line, start=tt_new_part.name[0], end=tt_new_part.name[-1],
        #     twiss_table=twiss_table,
        #     monitor_names_x=[nn for nn in monitor_names_x if nn in tt_new_part.name],
        #     monitor_names_y=[nn for nn in monitor_names_y if nn in tt_new_part.name],
        #     corrector_names_x=[nn for nn in corrector_names_x if nn in tt_new_part.name],
        #     corrector_names_y=[nn for nn in corrector_names_y if nn in tt_new_part.name],
        #     x_init=this_x_init, px_init=this_px_init,
        #     y_init=this_y_init, py_init=this_py_init,
        #     zeta_init=this_zeta_init, delta_init=this_delta_init,
        #     monitor_alignment=monitor_alignment
        # )
        # ocorr_only_added_part.correct(rcond=rcond_short, n_iter=1, verbose=False)

        # if verbose:
        #     ocprint = ocorr_only_added_part
        #     tw_orbit_print = ocprint.x_correction._compute_tw_orbit()
        #     x_meas_print = ocprint.x_correction._measure_position(tw_orbit_print)
        #     y_meas_print = ocprint.y_correction._measure_position(tw_orbit_print)
        #     str_2print = f'Stop at s={s_corr_end}, '
        #     str_2print += 'local rms  = ['
        #     str_2print += (f'x: {ocprint.x_correction._position_before.std():.2e}'
        #         f' -> {x_meas_print.std():.2e}, ')
        #     str_2print += (f'y: {ocprint.y_correction._position_before.std():.2e}'
        #         f' -> {y_meas_print.std():.2e}]')
        #     print(str_2print)

        # Correct from start line to end of new added portion
        tt_part = tt.rows[0:s_corr_end:'s']
        ocorr = TrajectoryCorrection(
            twiss_table=twiss_table,
            line=line, start=tt_part.name[0], end=tt_part.name[-1],
            monitor_names_x=[nn for nn in monitor_names_x if nn in tt_part.name],
            monitor_names_y=[nn for nn in monitor_names_y if nn in tt_part.name],
            corrector_names_x=[nn for nn in corrector_names_x if nn in tt_part.name],
            corrector_names_y=[nn for nn in corrector_names_y if nn in tt_part.name],
            x_init=x_init, px_init=px_init, y_init=y_init, py_init=py_init,
            zeta_init=zeta_init, delta_init=delta_init,
            monitor_alignment=monitor_alignment
        )
        ocorr.correct(rcond=rcond_long, n_iter=1, verbose=False)

        if verbose:
            ocprint = ocorr
            tw_orbit_print = ocprint.x_correction._compute_tw_orbit()
            x_meas_print = ocprint.x_correction._measure_position(tw_orbit_print)
            y_meas_print = ocprint.y_correction._measure_position(tw_orbit_print)
            str_2print = f'Stop at s={s_corr_end}, '
            str_2print += 'global rms = ['
            str_2print += (f'x: {ocprint.x_correction._position_before_iter.std():.2e}'
                f' -> {x_meas_print.std():.2e}, ')
            str_2print += (f'y: {ocprint.y_correction._position_before_iter.std():.2e}'
                f' -> {y_meas_print.std():.2e}]')
            print(str_2print)

        s_corr_end += ds_thread
        i_win += 1

    return ocorr



import xtrack as xt
_print = xt.general._print

class PipelineBranch:
    def __init__(self, line=None, particles=None, tracker=None):

        if tracker is not None:
            _print( 'Warning! '
                "The argument tracker is deprecated. Please use line instead.")
            assert line is None
            line = tracker.line

        if isinstance(line, xt.Tracker):
            _print('Warning! '
                "The use of Tracker as argument of `PipelineBranch` is deprecated."
                " Please use Line instead.")
            line = line.line

        self.line = line
        self.particles = particles
        self.pipeline_status = None

        if self.line.tracker is None:
            raise ValueError('Tracker not built for line')

        self.line.tracker.enable_pipeline_hold = True

class PipelineMultiTracker:

    def __init__(self, branches, enable_debug_log=False, verbose=False):
        self.branches = branches
        self.enable_debug_log = enable_debug_log
        self.verbose = verbose

        if self.enable_debug_log:
            self.debug_log = []

    def track(self, **kwargs):

        for branch in self.branches:
            branch.pipeline_status = branch.line.track(
                 branch.particles, **kwargs,
                 _called_by_pipeline=True)

        need_resume = True
        while need_resume:
            need_resume = False
            for i_branch, branch in enumerate(self.branches):
                if (branch.pipeline_status is not None
                        and branch.pipeline_status.on_hold):
                    if self.enable_debug_log:
                        self.debug_log.append({
                            'branch': i_branch,
                            'track_session_turn':
                                            branch.pipeline_status.data['tt'],
                            'held_by_element': branch.line.tracker._part_names[
                                            branch.pipeline_status.data['ipp']],
                            'info': branch.pipeline_status.data['status_from_element'].info
                        })
                    if self.verbose:
                        print(
                            f"Pipeline hold at branch {i_branch} "
                            f"at turn {branch.pipeline_status.data['tt']} "
                            f"by element {branch.line.tracker._part_names[branch.pipeline_status.data['ipp']]} "
                            f"with info: {branch.pipeline_status.data['status_from_element'].info}")

                    branch.pipeline_status = branch.line.tracker.resume(
                                                        branch.pipeline_status)
                    need_resume = True






import json
import pathlib

from cpymad.madx import Madx

import xobjects as xo
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
            __file__).parent.joinpath('../test_data').absolute()

correction_config = {
    'IR1 left': dict(
        ref_with_knobs={'on_corr_co': 0, 'on_disp': 0},
        start='e.ds.r8.b1',
        end='e.ds.l1.b1',
        vary=(
            'corr_co_acbh14.l1b1',
            'corr_co_acbh12.l1b1',
            'corr_co_acbv15.l1b1',
            'corr_co_acbv13.l1b1',
            ),
        targets=('e.ds.l1.b1',),
    ),
    'IR1 right': dict(
        ref_with_knobs={'on_corr_co': 0, 'on_disp': 0},
        start='s.ds.r1.b1',
        end='s.ds.l2.b1',
        vary=(
            'corr_co_acbh13.r1b1',
            'corr_co_acbh15.r1b1',
            'corr_co_acbv12.r1b1',
            'corr_co_acbv14.r1b1',
            ),
        targets=('s.ds.l2.b1',),
    ),
    'IR5 left': dict(
        ref_with_knobs={'on_corr_co': 0, 'on_disp': 0},
        start='e.ds.r4.b1',
        end='e.ds.l5.b1',
        vary=(
            'corr_co_acbh14.l5b1',
            'corr_co_acbh12.l5b1',
            'corr_co_acbv15.l5b1',
            'corr_co_acbv13.l5b1',
            ),
        targets=('e.ds.l5.b1',),
    ),
    'IR5 right': dict(
        ref_with_knobs={'on_corr_co': 0, 'on_disp': 0},
        start='s.ds.r5.b1',
        end='s.ds.l6.b1',
        vary=(
            'corr_co_acbh13.r5b1',
            'corr_co_acbh15.r5b1',
            'corr_co_acbv12.r5b1',
            'corr_co_acbv14.r5b1',
            ),
        targets=('s.ds.l6.b1',),
    ),
    'IP1': dict(
        ref_with_knobs={'on_corr_co': 0, 'on_disp': 0},
        start='e.ds.l1.b1',
        end='s.ds.r1.b1',
        vary=(
            'corr_co_acbch6.l1b1',
            'corr_co_acbcv5.l1b1',
            'corr_co_acbch5.r1b1',
            'corr_co_acbcv6.r1b1',
            'corr_co_acbyhs4.l1b1',
            'corr_co_acbyhs4.r1b1',
            'corr_co_acbyvs4.l1b1',
            'corr_co_acbyvs4.r1b1',
        ),
        targets=('ip1', 's.ds.r1.b1'),
    ),
    'IP2': dict(
        ref_with_knobs={'on_corr_co': 0, 'on_disp': 0},
        start='e.ds.l2.b1',
        end='s.ds.r2.b1',
        vary=(
            'corr_co_acbyhs5.l2b1',
            'corr_co_acbchs5.r2b1',
            'corr_co_acbyvs5.l2b1',
            'corr_co_acbcvs5.r2b1',
            'corr_co_acbyhs4.l2b1',
            'corr_co_acbyhs4.r2b1',
            'corr_co_acbyvs4.l2b1',
            'corr_co_acbyvs4.r2b1',
        ),
        targets=('ip2', 's.ds.r2.b1'),
    ),
    'IP5': dict(
        ref_with_knobs={'on_corr_co': 0, 'on_disp': 0},
        start='e.ds.l5.b1',
        end='s.ds.r5.b1',
        vary=(
            'corr_co_acbch6.l5b1',
            'corr_co_acbcv5.l5b1',
            'corr_co_acbch5.r5b1',
            'corr_co_acbcv6.r5b1',
            'corr_co_acbyhs4.l5b1',
            'corr_co_acbyhs4.r5b1',
            'corr_co_acbyvs4.l5b1',
            'corr_co_acbyvs4.r5b1',
        ),
        targets=('ip5', 's.ds.r5.b1'),
    ),
    'IP8': dict(
        ref_with_knobs={'on_corr_co': 0, 'on_disp': 0},
        start='e.ds.l8.b1',
        end='s.ds.r8.b1',
        vary=(
            'corr_co_acbch5.l8b1',
            'corr_co_acbyhs4.l8b1',
            'corr_co_acbyhs4.r8b1',
            'corr_co_acbyhs5.r8b1',
            'corr_co_acbcvs5.l8b1',
            'corr_co_acbyvs4.l8b1',
            'corr_co_acbyvs4.r8b1',
            'corr_co_acbyvs5.r8b1',
        ),
        targets=('ip8', 's.ds.r8.b1'),
    ),
}


@for_all_test_contexts
def test_orbit_correction(test_context):
    with open(test_data_folder /
              'hllhc14_no_errors_with_coupling_knobs/line_b1.json', 'r') as fid:
        dct_b1 = json.load(fid)

    input_line = xt.Line.from_dict(dct_b1)

    # Load line with knobs on correctors only
    mad = Madx(stdout=False)
    mad.call(str(test_data_folder /
                 'hllhc14_no_errors_with_coupling_knobs/lhcb1_seq.madx'))
    mad.use(sequence='lhcb1')
    input_line_co_ref = xt.Line.from_madx_sequence(
        mad.sequence.lhcb1,
        deferred_expressions=True,
        expressions_for_element_types=('kicker', 'hkicker', 'vkicker'),
    )

    collider = xt.Environment(lines={'lhcb1': input_line.copy()})
    collider.import_line(line_name='lhcb1_co_ref', line=input_line_co_ref.copy())
    collider['lhcb1_co_ref'].particle_ref = collider['lhcb1'].particle_ref.copy()
    collider.build_trackers(_context=test_context)

    # Wipe out orbit correction from pymask
    tt_corr = collider.vars.get_table().rows['corr_acb.*']
    for kk in tt_corr.name:
        collider.vars[kk] = 0

    collider.vars['on_x1'] = 0
    collider.vars['on_x2'] = 0
    collider.vars['on_x5'] = 0
    collider.vars['on_x8'] = 0
    collider.vars['on_sep1'] = 0
    collider.vars['on_sep2'] = 0
    collider.vars['on_sep5'] = 0
    collider.vars['on_sep8'] = 0
    collider.vars['on_lhcb'] = 0
    collider.vars['on_alice'] = 0

    # Check that in both machines the orbit is flat at the ips
    for nn in ['lhcb1', 'lhcb1_co_ref']:
        tw = collider[nn].twiss(method='4d', zeta0=0, delta0=0)
        for ip in ['ip1', 'ip2', 'ip5', 'ip8']:
            xo.assert_allclose(tw['x', ip], 0, 1e-10)
            xo.assert_allclose(tw['px', ip], 0, 1e-10)
            xo.assert_allclose(tw['y', ip], 0, 1e-10)
            xo.assert_allclose(tw['py', ip], 0, 1e-10)

    # Check that the tune knobs work only on line and not on line_co_ref
    tw0 = collider['lhcb1'].twiss(method='4d', zeta0=0, delta0=0)
    tw0_co_ref = collider['lhcb1_co_ref'].twiss(method='4d', zeta0=0, delta0=0)
    collider['lhcb1'].vars['kqtf.b1'] = 1e-5
    collider['lhcb1_co_ref'].vars['kqtf.b1'] = 1e-5 # This should not change anything
    tw1 = collider['lhcb1'].twiss(method='4d', zeta0=0, delta0=0)
    tw1_co_ref = collider['lhcb1_co_ref'].twiss(method='4d', zeta0=0, delta0=0)
    assert tw1.qx != tw0.qx
    assert tw1_co_ref.qx == tw0_co_ref.qx

    # Add correction term to all dipole correctors
    collider.vars['on_corr_co'] = 1
    for kk in list(collider.vars.keys()):
        if kk.startswith('acb'):
            collider.vars['corr_co_'+kk] = 0
            collider.vars[kk] += (collider.vars['corr_co_'+kk]
                                * collider.vars['on_corr_co'])

    # Set some orbit knobs in both machines and switch on experimental magnets
    collider.vars['on_x1'] = 250
    collider.vars['on_x2'] = 250
    collider.vars['on_x5'] = 250
    collider.vars['on_x8'] = 250
    collider.vars['on_disp'] = 1
    collider.vars['on_lhcb'] = 1
    collider.vars['on_alice'] = 1

    # Introduce dip kick in all triplets (only in line)
    collider['lhcb1']['mqxfb.b2l1..11'].knl[0] = 1e-6
    collider['lhcb1']['mqxfb.b2l1..11'].ksl[0] = 1.5e-6
    collider['lhcb1']['mqxfb.b2r1..11'].knl[0] = 2e-6
    collider['lhcb1']['mqxfb.b2r1..11'].ksl[0] = 1e-6
    collider['lhcb1']['mqxb.b2l2..11'].knl[0] = 1e-6
    collider['lhcb1']['mqxb.b2l2..11'].ksl[0] = 1.5e-6
    collider['lhcb1']['mqxb.b2r2..11'].knl[0] = 2e-6
    collider['lhcb1']['mqxb.b2r2..11'].ksl[0] = 1e-6
    collider['lhcb1']['mqxfb.b2l5..11'].knl[0] = 1e-6
    collider['lhcb1']['mqxfb.b2l5..11'].ksl[0] = 1.5e-6
    collider['lhcb1']['mqxfb.b2r5..11'].knl[0] = 2e-6
    collider['lhcb1']['mqxfb.b2r5..11'].ksl[0] = 1e-6
    collider['lhcb1']['mqxb.b2l8..11'].knl[0] = 1e-6
    collider['lhcb1']['mqxb.b2l8..11'].ksl[0] = 1.5e-6
    collider['lhcb1']['mqxb.b2r8..11'].knl[0] = 2e-6
    collider['lhcb1']['mqxb.b2r8..11'].ksl[0] = 1e-6

    tw_before = collider.lhcb1.twiss()

    collider.lhcb1._xmask_correct_closed_orbit(
                             reference=collider.lhcb1_co_ref,
                             correction_config=correction_config)

    tw = collider.lhcb1.twiss()

    xo.assert_allclose(tw['px', 'ip1'], 250e-6, rtol=5e-5, atol=1e-8)
    xo.assert_allclose(tw['py', 'ip1'], 0, atol=1e-8)

    xo.assert_allclose(tw['px', 'ip5'], 0, atol=1e-8)
    xo.assert_allclose(tw['py', 'ip5'], 250e-6, atol=1e-8)

    assert tw['px', 'ip2'] > 1e-7 # effect of the spectrometer tilt
    assert tw['py', 'ip2'] > 255e-6 # effect of the spectrometer
    xo.assert_allclose(tw['px', 'bpmsw.1r2.b1'], 0, atol=1e-8)  # external angle
    xo.assert_allclose(tw['py', 'bpmsw.1r2.b1'], 250e-6, atol=1e-8) # external angle

    assert tw['px', 'ip8'] > 255e-6 # effect of the spectrometer
    assert tw['py', 'ip8'] > 1e-6 # effect of the spectrometer tilt
    xo.assert_allclose(tw['px', 'bpmsw.1r8.b1'], 250e-6, atol=1e-8) # external angle
    xo.assert_allclose(tw['py', 'bpmsw.1r8.b1'], 0, atol=1e-8) # external angle


    places_to_check = [
    'e.ds.r8.b1',
    'e.ds.r4.b1',
    's.ds.l2.b1',
    's.ds.l6.b1',
    'e.ds.l1.b1',
    'e.ds.l2.b1',
    'e.ds.l5.b1',
    'e.ds.l8.b1',
    's.ds.r1.b1',
    's.ds.r2.b1',
    's.ds.r5.b1',
    's.ds.r8.b1']

    for place in places_to_check:
        xo.assert_allclose(tw['x', place], 0, atol=1e-6)
        xo.assert_allclose(tw['px', place], 0, atol=1e-8)
        xo.assert_allclose(tw['y', place], 0, atol=1e-6)
        xo.assert_allclose(tw['py', place], 0, atol=1e-8)

    with xt._temp_knobs(collider, dict(on_corr_co=0, on_disp=0)):
        tw_ref = collider.lhcb1_co_ref.twiss()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xobjects
import xpart
import xdeps
import xtrack
import xfields
import xmask
import xcoll
import xwakes

print("\n-------- Paths: --------")

print('xobjects: ', xobjects.__path__)
print('xpart:    ', xpart.__path__)
print('xdeps:    ', xdeps.__path__)
print('xtrack:   ', xtrack.__path__)
print('xfields:  ', xfields.__path__)
print('xmask     ', xmask.__path__)
print('xcoll     ', xcoll.__path__)
print('xwakes    ', xwakes.__path__)


try:
    from git import Repo
    from pathlib import Path

    print("\n-------- Git branches: --------")
    repo_xobjects = Repo(Path(xobjects.__path__[0]).parent)
    repo_xpart = Repo(Path(xpart.__path__[0]).parent)
    repo_xdeps = Repo(Path(xdeps.__path__[0]).parent)
    repo_xtrack = Repo(Path(xtrack.__path__[0]).parent)
    repo_xfields = Repo(Path(xfields.__path__[0]).parent)
    repo_xmask = Repo(Path(xmask.__path__[0]).parent)
    repo_xcoll = Repo(Path(xcoll.__path__[0]).parent)
    repo_xwakes = Repo(Path(xwakes.__path__[0]).parent)

    print('xobjects is on branch: ', repo_xobjects.active_branch)
    print('xpart is on branch:    ', repo_xpart.active_branch)
    print('xdeps is on branch:    ', repo_xdeps.active_branch)
    print('xtrack is on branch:   ', repo_xtrack.active_branch)
    print('xfields is on branch:  ', repo_xfields.active_branch)
    print('xmask is on branch:    ', repo_xmask.active_branch)
    print('xcoll is on branch:    ', repo_xcoll.active_branch)
    print('xwakes is on branch:   ', repo_xwakes.active_branch)

except Exception as err:
    print('No git info because of the following exception: ', err)

print("\n-------- Enabled test contexts: --------")
for cc in xobjects.context.get_test_contexts():
        print(cc)



import time

import xtrack as xt
import xobjects as xo

ctx=xo.ContextCpu()

def mkbend(length=1,angle=0.1,steps=60):
    dw=[ 1/(steps+1) for _ in range(steps+1)]
    kw=[ 1/steps for _ in range(steps)]
    el=[xt.Drift(length=length*dw[0])]
    for dd,kk in zip(dw,kw):
        el.append(xt.Multipole(knl=[angle*kk],hxl=angle*kk,length=length*kk))
        el.append(xt.Drift(length=length*dd))
    return el

t1 = time.perf_counter()
for step in range(1,20):

    elements=mkbend(1,0.1,step)
    line=xt.Line(elements,
                 particle_ref=xt.Particles(mass0=xt.ELECTRON_MASS_EV,p0c=100e9,_context=ctx))
    line.build_tracker(_context=ctx)
    line.configure_radiation(model='mean')
    pp=line.build_particles()
    line.track(pp)
    print(step)
    print(f"px = {pp.px[0]}")
    print(f"py = {pp.py[0]}")
    print(f"pt = {pp.ptau[0]}")
t2 = time.perf_counter()
print(f"Time: {t2-t1}")

elements=mkbend(1,0.1, 2)
line=xt.Line(elements,
                particle_ref=xt.Particles(mass0=xt.ELECTRON_MASS_EV,p0c=100e9,_context=ctx))
line.build_tracker(_context=ctx)
line.configure_radiation(model='mean')

from xtrack.prebuild_kernels import get_suitable_kernel
get_suitable_kernel(
    config=line.config,
    line_element_classes=line.tracker.line_element_classes,
    verbose=True)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #


def make_short_line(line):
    new_elements = []
    new_names = []
    found_types = []
    for ee, nn in zip(line.elements, line.element_names):
        if ee.__class__ not in found_types:
            new_elements.append(ee)
            new_names.append(nn)
            found_types.append(ee.__class__)
    line.elements = new_elements
    line.element_names = new_names

    return line



import xtrack as xt
import numpy as np
import pytest

from xobjects.test_helpers import for_all_test_contexts
import xobjects as xo

assert_allclose = np.testing.assert_allclose

slice_mode = 'thin'
tilted = True
orientation = 'acw'
transform_to_actual_elements = True

if hasattr(np, 'trapezoid'): # numpy >= 2.0
    trapz = np.trapezoid
else:
    trapz = np.trapz

@for_all_test_contexts
@pytest.mark.parametrize(
    'slice_mode',
    [None, 'thin', 'thick'],
    ids=['no_slice', 'thin_slice', 'thick_slice'])
@pytest.mark.parametrize(
    'tilted',
    [False, True],
    ids=['not_tilted', 'tilted'])
@pytest.mark.parametrize(
    'orientation',
    ['cw', 'acw'],
    ids=['cw', 'acw'])
@pytest.mark.parametrize(
    'transform_to_actual_elements',
    [False, True],
    ids=['no_actual_elements', 'actual_elements'])
def test_survey_slicing(test_context, slice_mode, tilted, orientation,
                        transform_to_actual_elements):

    line = xt.Line(
        elements=[
            xt.Drift(),
            xt.Bend(),
            xt.Drift(),
            xt.Bend(),
            xt.Drift(),
            xt.Bend(),
            xt.Drift(),
            xt.Bend(),
        ]
    )
    line.build_tracker(_context=test_context)

    line.vars['l_drift'] = 999.
    line.vars['l_bend'] = 999.
    line.vars['h_bend'] = 999.
    line.vars['tilt_bend_deg'] = 999.


    line.element_refs['e0'].length = line.vars['l_drift']
    line.element_refs['e1'].length = line.vars['l_bend']
    line.element_refs['e2'].length = line.vars['l_drift']
    line.element_refs['e3'].length = line.vars['l_bend']
    line.element_refs['e4'].length = line.vars['l_drift']
    line.element_refs['e5'].length = line.vars['l_bend']
    line.element_refs['e6'].length = line.vars['l_drift']
    line.element_refs['e7'].length = line.vars['l_bend']

    line.element_refs['e1'].h = line.vars['h_bend']
    line.element_refs['e3'].h = line.vars['h_bend']
    line.element_refs['e5'].h = line.vars['h_bend']
    line.element_refs['e7'].h = line.vars['h_bend']

    line.element_refs['e1'].rot_s_rad = line.vars['tilt_bend_deg'] * np.pi / 180
    line.element_refs['e3'].rot_s_rad = line.vars['tilt_bend_deg'] * np.pi / 180
    line.element_refs['e5'].rot_s_rad = line.vars['tilt_bend_deg'] * np.pi / 180
    line.element_refs['e7'].rot_s_rad = line.vars['tilt_bend_deg'] * np.pi / 180

    if slice_mode is not None:
        line.slice_thick_elements(
            slicing_strategies=[xt.Strategy(xt.Teapot(20, mode=slice_mode))])
        line.build_tracker(_context=test_context)

    line.vars['l_drift'] = 1
    line.vars['l_bend'] = 1
    if orientation == 'cw':
        line.vars['h_bend'] = np.pi/2 / line.vars['l_bend']
    elif orientation == 'acw':
        line.vars['h_bend'] = -np.pi/2 / line.vars['l_bend']

    if tilted:
        line.vars['tilt_bend_deg'] = 90
    else:
        line.vars['tilt_bend_deg'] = 0

    if slice_mode == 'thin' and transform_to_actual_elements:
        line.discard_tracker()
        line._replace_with_equivalent_elements()
        line.build_tracker(_context=test_context)
        assert isinstance(line['e1..1'], xt.Multipole)

    sv = line.survey()
    assert_allclose(sv.Z[-1], 0, rtol=0, atol=1e-13)
    assert_allclose(sv.X[-1], 0, rtol=0, atol=1e-13)
    assert_allclose(sv.Y[-1], 0, rtol=0, atol=1e-13)
    assert_allclose(sv.s[-1], 8, rtol=0, atol=1e-13)


    if not tilted and orientation == 'acw':
        assert_allclose(np.abs(sv.Y), 0, rtol=0, atol=1e-14)
        assert_allclose(trapz(sv.X, sv.Z), -4.818 , rtol=0, # anti-clockwise
                        atol=(2e-3 if slice_mode is not None else 0.5))
    elif not tilted and orientation == 'cw':
        assert_allclose(np.abs(sv.Y), 0, rtol=0, atol=1e-14)
        assert_allclose(trapz(sv.X, sv.Z), 4.818 , rtol=0, # clockwise
                        atol=(2e-3 if slice_mode is not None else 0.5))
    elif tilted and orientation == 'acw':
        assert_allclose(np.abs(sv.X), 0, rtol=0, atol=1e-14)
        assert_allclose(trapz(sv.Y, sv.Z), -4.818 , rtol=0, # anti-clockwise
                        atol=(2e-3 if slice_mode is not None else 0.5))
    elif tilted and orientation == 'cw':
        assert_allclose(np.abs(sv.X), 0, rtol=0, atol=1e-14)
        assert_allclose(trapz(sv.Y, sv.Z), 4.818 , rtol=0, # clockwise
                        atol=(2e-3 if slice_mode is not None else 0.5))



import numpy as np
import xtrack as xt
import xobjects as xo
import xpart as xp

import pytest

def test_rf_track_lattice():

    try:
        import RF_Track as RFT
    except ImportError:
        pytest.skip()

    # Bend parameters
    clight = 299792458 # m/s
    p0c = 1.2e9  # eV
    lbend = 3 # m
    angle = np.pi / 2 # rad
    rho = lbend / angle # m
    By = p0c / rho / clight # T

    #############################################
    #######  RF-Track's part starts here  #######
    #############################################

    # Define the RF-Track element
    vol = RFT.Volume()
    vol.dt_mm = 0.1
    vol.odeint_algorithm = 'rk2'
    vol.set_static_Bfield(0.0, By, 0.0)
    vol.set_s0(rho, 0.0, 0.0, 0.0, 0.0, 0.0)
    vol.set_s1(0.0, 0.0, rho, 0.0, 0.0, -angle)
    vol.set_length(lbend)

    #############################################
    #######  RF-Track's part ends here    #######
    #############################################

    # Back to Xsuite
    pi = np.pi
    elements = {
        'd1.1':  xt.Drift(length=1),
        'mb1.1': xt.RFT_Element(element=vol),
        'd2.1':  xt.Drift(length=1),

        'mqd.1': xt.Quadrupole(length=0.3, k1=-0.7),
        'd3.1':  xt.Drift(length=1),
        'mb2.1': xt.RFT_Element(element=vol),
        'd4.1':  xt.Drift(length=1),

        'd1.2':  xt.Drift(length=1),
        'mb1.2': xt.RFT_Element(element=vol),
        'd2.2':  xt.Drift(length=1),

        'mqd.2': xt.Quadrupole(length=0.3, k1=-0.7),
        'd3.2':  xt.Drift(length=1),
        'mb2.2': xt.RFT_Element(element=vol),
        'd4.2':  xt.Drift(length=1),
    }

    # Build the ring
    line = xt.Line(elements=elements, element_names=list(elements.keys()))
    line.particle_ref = xt.Particles(p0c=p0c, mass0=xt.PROTON_MASS_EV)
    line.configure_bend_model(core='full', edge=None)


    ## Transfer lattice on context and compile tracking code=
    line.build_tracker()

    ## Build particle object on context
    n_part = 200

    rng = np.random.default_rng(2021)

    particles0 = xp.Particles(p0c=p0c, #eV
                            q0=1, mass0=xp.PROTON_MASS_EV,
                            x=rng.uniform(-1e-3, 1e-3, n_part),
                            px=rng.uniform(-1e-5, 1e-5, n_part),
                            y=rng.uniform(-2e-3, 2e-3, n_part),
                            py=rng.uniform(-3e-5, 3e-5, n_part),
                            zeta=rng.uniform(-1e-2, 1e-2, n_part),
                            delta=rng.uniform(-1e-2, 1e-2, n_part))

    particles_rft = particles0.copy()

    print('RF-Track tracking starts')
    line.track(particles_rft)   
    print('RF-Track tracking ends')

    ele_xt = elements.copy()
    b_xt = xt.Bend(length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend)
    ele_xt['mb1.1'] = b_xt
    ele_xt['mb1.2'] = b_xt
    ele_xt['mb2.1'] = b_xt
    ele_xt['mb2.2'] = b_xt

    line_ref = xt.Line(elements=ele_xt, element_names=line.element_names)
    line_ref.build_tracker()
    particles_ref = particles0.copy()

    print('Xtrack tracking starts')
    line_ref.track(particles_ref)
    print('Xtrack tracking ends')

    assert_allclose = np.testing.assert_allclose
    assert np.all(particles_rft.particle_id == particles_ref.particle_id)
    assert_allclose(particles_rft.x,  particles_ref.x,  atol=6e-5, rtol=0)
    assert_allclose(particles_rft.px, particles_ref.px, atol=2e-5, rtol=0)
    assert_allclose(particles_rft.y,  particles_ref.y,  atol=6e-7, rtol=0)
    assert_allclose(particles_rft.py, particles_ref.py, atol=2e-6, rtol=0)
    assert_allclose(particles_rft.rpp, particles_ref.rpp, atol=3e-5, rtol=0)
    assert_allclose(particles_rft.rvv, particles_ref.rvv, atol=1e-5, rtol=0)
    assert_allclose(particles_rft.ptau, particles_ref.ptau, atol=2e-5, rtol=0)
    assert_allclose(particles_rft.delta, particles_ref.delta, atol=4e-5, rtol=0)
    assert_allclose(particles_rft.chi, particles_ref.chi, atol=1e-10, rtol=0)
    assert_allclose(particles_rft.p0c, particles_ref.p0c, atol=4e-5, rtol=0)
    assert_allclose(particles_rft.energy0, particles_ref.energy0, atol=4e-5, rtol=0)
    assert_allclose(particles_rft.zeta, particles_ref.zeta, atol=4e-5, rtol=0)
    assert_allclose(particles_rft.beta0, particles_ref.beta0, atol=1e-10, rtol=0)
    assert_allclose(particles_rft.mass0, particles_ref.mass0, atol=1e-10, rtol=0)
    assert_allclose(particles_rft.gamma0, particles_ref.gamma0, atol=1e-10, rtol=0)



import numpy as np

from xtrack import PipelineStatus

class DummyPipelinedElement:
    def __init__(self, n_hold=0):
        self.n_hold = n_hold
        self.iscollective = True
        self.i_hold = 0

    def track(self, particles):
        self.i_hold += 1
        if self.i_hold < self.n_hold:
            return PipelineStatus(on_hold=True,
                info=f'stopped by internal counter {self.i_hold}/{self.n_hold}')
        else:
            self.i_hold = 0


import xtrack as xt

line=xt.Line(elements=[xt.Drift(length=1),
                           DummyPipelinedElement(n_hold=3),
                           xt.Drift(length=1)])
line.build_tracker(enable_pipeline_hold=True)


p = xt.Particles(p0c=7e12, x=[0,0,0])

session_on_hold = line.track(p, num_turns=2)
assert np.all(p.s==1)
assert np.all(p.at_turn==0)
session_on_hold = line.tracker.resume(session_on_hold)
assert np.all(p.s==1)
assert np.all(p.at_turn==0)
session_on_hold = line.tracker.resume(session_on_hold)
assert np.all(p.s==1)
assert np.all(p.at_turn==1)


import numpy as np
from cpymad.madx import Madx
import xtrack as xt

mad = Madx()
mad.input("""
beam, particle=proton, pc = 14.0;
BRHO      = BEAM->PC * 3.3356;
""")
mad.call("../../test_data/ps_sftpro/ps.seq")
mad.call("../../test_data/ps_sftpro/ps_hs_sftpro.str")
mad.use('ps')
twm = mad.twiss()

line = xt.Line.from_madx_sequence(mad.sequence.ps, allow_thick=True,
                                  deferred_expressions=True,
                                  )
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV,
                                    q0=1, gamma0=mad.sequence.ps.beam.gamma)
line.twiss_default['method'] = '4d'

tw = line.twiss()

opt = line.match(
    solve=False,
    vary=[xt.VaryList(['kf', 'kd'], step=1e-5)],
    targets=[xt.TargetSet(qx=6.255278, qy=6.29826, tol=1e-7)],
)
opt.solve()

r0 = np.linspace(0, 100, 50)
p = line.build_particles(
    x_norm=r0*np.cos(np.pi/20.),
    px_norm=r0*np.sin(np.pi/20.),
    nemitt_x=1e-6, nemitt_y=1e-6)

line.track(p, num_turns=1000, turn_by_turn_monitor=True, with_progress=True)
mon = line.record_last_track

tw_mt = line.twiss(co_guess={'x': 0.025}, num_turns=4)
tw_core = line.twiss(co_guess={'x': 0.0}, num_turns=0)

# Inspect and plot
tw_start_turns = tw_mt.rows['_turn_.*']
tw_start_turns.show()
import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(mon.x.flatten(), mon.px.flatten(), '.', markersize=1)
plt.plot(tw_start_turns.x, tw_start_turns.px, '*')
plt.ylim(-0.004, 0.004)
plt.xlim(-0.08, 0.08)

plt.figure(2)
ax1 = plt.subplot(3,1,1)
plt.plot(tw_mt.s, tw_mt.x)
plt.plot(tw_mt.s, tw_mt.y)
plt.subplot(3,1,2, sharex=ax1)
plt.plot(tw_mt.s, tw_mt.betx)
plt.plot(tw_core.s, tw_core.betx)
plt.subplot(3,1,3, sharex=ax1)
plt.plot(tw_mt.s, tw_mt.bety)
plt.plot(tw_core.s, tw_core.bety)

plt.show()

assert len(tw_mt.rows['_turn.*']) == 4
assert len(tw_mt.rows['_end_poi.*']) == 1

assert '_turn_0' in tw_mt.name
assert '_turn_1' in tw_mt.name
assert '_turn_2' in tw_mt.name
assert '_turn_3' in tw_mt.name

assert np.all(np.diff(tw_mt.s) >= 0)

circum = line.get_length()

assert np.isclose(tw_mt.s[-1], 4 * circum, atol=1e-10, rtol=0)
assert np.isclose(tw_mt['s', '_turn_0'], 0, atol=1e-10, rtol=0)
assert np.isclose(tw_mt['s', '_turn_1'], circum, atol=1e-10, rtol=0)
assert np.isclose(tw_mt['s', '_turn_2'], 2 * circum, atol=1e-10, rtol=0)
assert np.isclose(tw_mt['s', '_turn_3'], 3 * circum, atol=1e-10, rtol=0)
assert np.isclose(tw_mt['s', '_end_point'], 4 * circum, atol=1e-10, rtol=0)

assert np.isclose(tw_mt.mux[-1], 4 * tw.mux[-1], rtol=0, atol=0.05)
assert np.isclose(tw_mt.muy[-1], 4 * tw.muy[-1], rtol=0, atol=0.05)

assert 'qx' in tw
assert 'qy' in tw
assert 'qx' not in tw_mt
assert 'qy' not in tw_mt

assert tw_mt['x', '_turn_0'] > 2e-2
assert np.abs(tw_mt['x', '_turn_1']) < 1e-2
assert tw_mt['x', '_turn_2'] < -2e-2
assert np.abs(tw_mt['x', '_turn_3']) < 1e-2

assert np.isclose(tw_mt.x[-1], tw_mt.x[0], rtol=0, atol=1e-8)
assert np.isclose(tw_mt.y[-1], tw_mt.y[0], rtol=0, atol=1e-8)
assert np.isclose(tw_mt.px[-1], tw_mt.px[0], rtol=0, atol=1e-10)
assert np.isclose(tw_mt.py[-1], tw_mt.py[0], rtol=0, atol=1e-10)
assert np.isclose(tw_mt.betx[-1], tw_mt.betx[0], rtol=0, atol=1e-5)
assert np.isclose(tw_mt.bety[-1], tw_mt.bety[0], rtol=0, atol=1e-5)
assert np.isclose(tw_mt.alfx[-1], tw_mt.alfx[0], rtol=0, atol=1e-5)
assert np.isclose(tw_mt.alfy[-1], tw_mt.alfy[0], rtol=0, atol=1e-5)